from enum import Enum


class StepBaseEnum(str, Enum):
    # general settings
    STEPID = "step_id"

    # different step types
    STEP_TYPE = "type"
    STEP_CREST = "CREST"
    STEP_OMEGA = "OMEGA"
    STEP_XTB = "XTB"
    STEP_MACROMODEL = "MACROMODEL"
    STEP_TURBOMOLE = "TURBOMOLE"
    STEP_COSMO = "COSMO"
    STEP_INITIALIZATION = "INITIALIZATION"
    STEP_EMBEDDING = "EMBEDDING"
    STEP_PREDICTION = "PREDICTION"
    STEP_MODEL_BUILDING = "MODEL_BUILDING"
    STEP_FEATURE_COUNTER = "FEATURE_COUNTER"
    STEP_BOLTZMANN_WEIGHTING = "BOLTZMANN_WEIGHTING"
    STEP_PKA_PREDICTION = "PKA_PREDICTION"
    STEP_PRIME = "PRIME"
    STEP_CLUSTERING = "CLUSTERING"
    STEP_RMSD = "RMSD"
    STEP_RMSFILTER = "RMSFILTER"
    STEP_DATA_MANIPULATION = "DATA_MANIPULATION"
    STEP_DESMOND = "DESMOND"
    STEP_DESMOND_SETUP = "DESMOND_SETUP"
    STEP_FILTER = "FILTER"
    STEP_PANTHER = "PANTHER"
    STEP_KALLISTO = "KALLISTO"
    STEP_JAZZY = "JAZZY"
    STEP_SHAEP = "SHAEP"
    STEP_PDB2GMX = "PDB2GMX"
    STEP_EDITCONF = "EDITCONF"
    STEP_SOLVATE = "SOLVATE"
    STEP_GENION = "GENION"
    STEP_GROMPP = "GROMPP"
    STEP_MDRUN = "MDRUN"
    STEP_TRJCONV = "TRJCONV"
    STEP_TRJCAT = "TRJCAT"
    STEP_GMX_RMSD = "GMX_RMSD"
    STEP_CLUSTER = "CLUSTER"
    STEP_DO_DSSP = "DO_DSSP"
    STEP_LIGPREP = "LIGPREP"
    STEP_GLIDE = "GLIDE"
    STEP_RESIDUE_SCANNING = "RESIDUE_SCANNING"
    STEP_AUTODOCKVINA_DOCKING = "VINA_DOCKING"
    STEP_AUTODOCKVINA_TARGET_PREPARATION = "VINA_TARGET_PREPARATION"
    STEP_GOLD_DOCKING = "GOLD_DOCKING"
    STEP_FEP_PLUS_SETUP = "FEP_PLUS_SETUP"
    STEP_FEP_PLUS_EXEC = "FEP_PLUS_EXEC"
    STEP_FEP_PLUS_ANALYSIS = "FEP_PLUS_ANALYSIS"
    STEP_FEP_ABSOLUTE = "FEP_ABSOLUTE"
    STEP_PREPWIZARD = "PREPWIZARD"
    STEP_MDPOCKET = "MDPOCKET"
    STEP_PDB_FIXER = "PDB_FIXER"
    STEP_PEPTIDE_EMBEDDER = "PEPTIDE_EMBEDDER"
    STEP_GMX_MMPBSA = "GMX_MMPBSA"

    # PMX SCRIPTS
    STEP_PMX_ABFE_SETUP = "PMX_ABFE_SETUP"
    STEP_PMX_ANALYSE = "PMX_ANALYSE"
    STEP_PMX_ATOMMAPPING = "PMX_ATOMMAPPING"
    STEP_PMX_DOUBLEBOX = "PMX_DOUBLEBOX"
    STEP_PMX_GENLIB = "PMX_GENLIB"
    STEP_PMX_GENTOP = "PMX_GENTOP"
    STEP_PMX_LIGANDHYBRID = "PMX_LIGANDHYBRID"
    STEP_PMX_MUTATE = "PMX_MUTATE"
    STEP_PMX_SETUP = "PMX_SETUP"
    STEP_PMX_PREPARE_SIMULATIONS = "PMX_PREPARE_SIMULATIONS"
    STEP_PMX_BOX_WATER_IONS = "PMX_BOX_WATER_IONS"
    STEP_PMX_PREPARE_TRANSITIONS = "PMX_PREPARE_TRANSITIONS"
    STEP_PMX_RUN_SIMULATIONS = "PMX_RUN_SIMULATIONS"
    STEP_PMX_ASSEMBLE_SYSTEMS = "PMX_ASSEMBLE_SYSTEMS"
    STEP_PMX_RUN_ANALYSIS = "PMX_RUN_ANALYSIS"

    STEP_PROTEIN_INTERACTIONS = "PROTEIN_INTERACTION"
    STEP_LOMAP = "LOMAP"

    STEP_DSSP = "DSSP"
    STEP_TS_CLUSTER = "TS_CLUSTER"
    STEP_ESP_SIM = "ESP_SIM"
    STEP_DISPATCHER = "DISPATCHER"
    STEP_ACTIVE_LEARNING = "ACTIVE_LEARNING"
    STEP_PROSPECTIVE_REINVENT = "PROSPECTIVE_REINVENT"

    # flow control blocks
    STEP_ITERATOR = "ITERATOR"

    # execution
    EXEC = "execution"
    EXEC_PREFIXEXECUTION = "prefix_execution"
    EXEC_BINARYLOCATION = "binary_location"
    EXEC_PARALLELIZATION = "parallelization"
    EXEC_PARALLELIZATION_CORES = "jobs"
    EXEC_PARALLELIZATION_MAXLENSUBLIST = "max_length_sublists"
    EXEC_FAILUREPOLICY = "failure_policy"
    EXEC_FAILUREPOLICY_NTRIES = "n_tries"
    EXEC_PLATFORM = "platform"
    EXEC_RESOURCES = "resources"
    EXEC_RESOURCES_PARTITION = "partition"
    EXEC_RESOURCES_GRES = "gres"
    EXEC_RESOURCES_TASKS = "tasks"
    EXEC_RESOURCES_MODULES = "modules"
    EXEC_RESOURCES_MEM = "mem"
    EXEC_RESOURCES_CORES = "cores"
    EXEC_RESOURCES_OTHER_ARGS = "other_args"
    EXEC_RESOURCES_ADDITIONAL_LINES = "additional_lines"

    # settings
    SETTINGS = "settings"
    SETTINGS_ARGUMENTS = "arguments"
    SETTINGS_ARGUMENTS_FLAGS = "flags"
    SETTINGS_ARGUMENTS_PARAMETERS = "parameters"
    SETTINGS_ADDITIONAL = "additional"

    PIPE_INPUT = "pipe_input"

    # annotation: fixed strings
    ANNOTATION_TAG_DOCKING_SCORE = "docking_score"
    ANNOTATION_TAG_G_SCORE = "g_score"

    ANNOTATION_GRID_ID = "grid_id"
    ANNOTATION_GRID_PATH = "grid_path"
    ANNOTATION_GRID_FILENAME = "grid_filename"

    GRID_IDS = "grid_ids"  # enforces given list of IDs rather than indices

    # I/O and "hand-over"
    # ---------
    FORMAT_SDF = "SDF"
    FORMAT_CSV = "CSV"
    FORMAT_SMI = "SMI"
    FORMAT_MOL2 = "MOL2"
    FORMAT_XTB = "XTB"
    FORMAT_PDB = "PDB"
    FORMAT_PKL = "PKL"
    FORMAT_SMILES = "SMILES"
    FORMAT_PLAIN = "PLAIN"
    FORMAT_TXT = "TXT"
    FORMAT_JSON = "JSON"
    FORMAT_DTR = "DTR"
    FORMAT_CMS = "CMS"

    INPUT = "input"
    INPUT_FIELD = "field"
    INPUT_SOURCES = "sources"
    INPUT_COMPOUNDS = "compounds"
    INPUT_ENUMERATIONS = "enumerations"
    INPUT_EXTENSION = "extension"
    INPUT_SOURCE = "source"
    INPUT_GENERIC = "generic"
    INPUT_GMX = "gmx_state"
    INPUT_FORMAT = "format"
    INPUT_SOURCE_TYPE = "source_type"
    INPUT_SOURCE_TYPE_FILE = "file"
    INPUT_SOURCE_TYPE_DIR = "dir"
    INPUT_SOURCE_TYPE_PATH = "path"
    INPUT_SOURCE_TYPE_STEP = "step"
    INPUT_SOURCE_TYPE_STRING = "string"
    INPUT_SOURCE_TYPE_URL = "url"

    INPUT_ENFORCE_IDS = "enforce_ids"
    INPUT_ENFORCE_COMPOUND_IDS = "compound_ids"
    INPUT_ENFORCE_ENUMERATION_IDS = "enumeration_ids"

    INPUT_MERGE = "merge"
    INPUT_MERGE_COMPOUNDS = "compounds"
    INPUT_MERGE_COMPOUNDS_BY = "merge_compounds_by"
    INPUT_MERGE_ENUMERATIONS = "enumerations"
    INPUT_MERGE_ENUMERATIONS_BY = "merge_enumerations_by"
    INPUT_MERGE_BY_NAME = "name"
    INPUT_MERGE_BY_ID = "id"
    INPUT_MERGE_BY_SMILE = "smile"

    FILE_TYPE_PDB = "pdb"
    FILETYPE_TXT = "txt"
    FILE_SIZE_THRESHOLD = 2e9

    # CSV settings
    INPUT_CSV_DELIMITER = "delimiter"
    INPUT_CSV_COLUMNS = "columns"
    INPUT_CSV_SMILES_COLUMN = "smiles"
    INPUT_CSV_NAMES_COLUMN = "names"

    # write-out settings
    WRITEOUT = "writeout"
    WRITEOUT_CONFIG = "config"

    WRITEOUT_COMP = "compounds"
    WRITEOUT_COMP_CATEGORY = "category"
    WRITEOUT_COMP_CATEGORY_CONFORMERS = "conformers"
    WRITEOUT_COMP_CATEGORY_ENUMERATIONS = "enumerations"
    WRITEOUT_COMP_CATEGORY_EXTRADATA = "extra_data"
    WRITEOUT_COMP_KEY = "key"
    WRITEOUT_COMP_AGGREGATION = "aggregation"
    WRITEOUT_COMP_AGGREGATION_MODE = "mode"
    WRITEOUT_COMP_AGGREGATION_MODE_ALL = "all"
    WRITEOUT_COMP_AGGREGATION_MODE_BESTPERCOMPOUND = "best_per_compound"
    WRITEOUT_COMP_AGGREGATION_MODE_BESTPERENUMERATION = "best_per_enumeration"
    WRITEOUT_COMP_AGGREGATION_HIGHESTISBEST = "highest_is_best"
    WRITEOUT_COMP_AGGREGATION_KEY = "key"
    WRITEOUT_COMP_SELECTED_TAGS = "selected_tags"
    WRITEOUT_COMP_SELECTED_TAGS_KEY = "key"
    WRITEOUT_COMP_SELECTED_TAGS_HIGHESTISBEST = "highest_is_best"

    WRITEOUT_GENERIC = "generic"
    WRITEOUT_GENERIC_KEY = "key"

    WRITEOUT_GMX = "gmx_state"

    WRITEOUT_DESTINATION = "destination"
    WRITEOUT_DESTINATION_RESOURCE = "resource"
    WRITEOUT_DESTINATION_TYPE = "type"
    WRITEOUT_DESTINATION_TYPE_FILE = "file"
    WRITEOUT_DESTINATION_TYPE_REINVENT = "reinvent"
    WRITEOUT_DESTINATION_TYPE_STDOUT = "stdout"
    WRITEOUT_DESTINATION_TYPE_STDERR = "stderr"
    WRITEOUT_DESTINATION_TYPE_REST = "rest"
    WRITEOUT_DESTINATION_FORMAT = "format"
    WRITEOUT_DESTINATION_MERGE = "merge"
    WRITEOUT_DESTINATION_AUTOMATIC = "automatic"
    WRITEOUT_DESTINATION_BASE_NAME = "base_name"
    WRITEOUT_DESTINATION_DIR = "dir"
    WRITEOUT_DESTINATION_MODE = "mode"

    TOKEN_GUARD = "token_guard"

    # try to find the internal value and return
    # def __getattr__(self, name):
    #     if name in self:
    #         return name
    #     raise AttributeError

    # prohibit any attempt to set any values
    # def __setattr__(self, key, value):
    #     raise ValueError("No changes allowed.")


class IteratorEnum(str, Enum):
    N_ITERS = "n_iters"
    ALL = "all"
    SINGLE = "single"


class StepEmbeddingEnum:
    METHOD = "method"
    METHOD_RDKIT = "RDKIT"

    EMBED_AS = "embed_as"
    EMBED_AS_ENUMERATIONS = "enumerations"
    EMBED_AS_CONFORMERS = "conformers"

    RDKIT_PROTONATE = "protonate"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepClusteringEnum:
    N_CLUSTERS = "n_clusters"
    MAX_ITER = "max_iter"
    TOP_N_PER_SOLVENT = "top_n_per_solvent"
    FEATURES = "features"
    FREE_ENERGY_SOLVENT_TAGS = "free_energy_solvent_tags"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepRMSFilterEnum:
    THRESHOLD = "threshold"  # RMS threshold in Angstrom

    # order by this tag in picking the conformers
    ORDER_BY = "order_by"
    ORDER_ASCENDING = "ascending"

    METHOD = "method"  # calculation method
    METHOD_BEST = "best"  # RDkit's "GetBestRMS"
    METHOD_ALIGNMOL = "alignmol"  # RDkit's "AlignMol"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepRMSDEnum:

    METHOD = "method"  # calculation method
    METHOD_BEST = "best"  # RDkit's "GetBestRMS"
    METHOD_ALIGNMOL = "alignmol"  # RDkit's "AlignMol"

    RMSD_TAG = "rmsd"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepFeatureCounterEnum:

    FEATURE = "feature"
    LEVEL = "level"
    LEVEL_ENUMERATION = "enumeration"
    LEVEL_CONFORMER = "conformer"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepDataManipulationEnum:

    # specify actions that can be used
    ACTION = "action"
    ACTION_NO_ACTION = (
        # used to skip any calculation (e.g. to just pool input data)
        "no_action"
    )
    CONVERT_MAE_TO_PDB = "mae2pdb"
    ASSEMBLE_COMPLEXES = "assemble_complexes"
    ACTION_ATTACH_CONFORMERS_AS_EXTRA = "attach_conformers_as_extra"
    COLLECT_ITERATOR_RESULTS = "collect_iterator_results"
    COMPUTE_MCS = "compute_mcs"
    FILTER = "filter"

    # --> For ACTION_ATTACH_CONFORMERS_AS_EXTRA
    # --- Match everything with the same <compound_id>:<enumeration_id>:<conformer_id> string
    MATCH_SOURCE = (
        "source"  # step from which the conformers are to be used for matching
    )
    KEY_MATCHED = "matched"  # extra-data key for matched data
    RECEPTOR = "receptor"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepBoltzmannWeightingEnum:
    PROPERTIES = "properties"
    PROPERTIES_INPUT = "input"
    PROPERTIES_OUTPUT = "output"

    WEIGHT = "weight"
    WEIGHT_INPUT = "input"
    WEIGHT_OUTPUT_PREFIX = "output_prefix"
    WEIGHT_PROPERTIES = "properties"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepPredictorEnum:
    MODEL_PATH = "model_path"
    FEATURES = "features"
    NAME_PREDICTED = "name_predicted"
    NAME_PREDICTED_DEFAULT = "pred_value"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepGoldEnum:

    CONFIGURATION = "configuration"
    GOLD_CONFIG_FILE = "gold_config_file"
    BLOCK_INDENT = "  "
    CONFIGURATION_START = "GOLD CONFIGURATION FILE"
    AUTOMATIC_SETTINGS = "AUTOMATIC SETTINGS"
    AUTOSCALE = "autoscale"
    POPULATION = "POPULATION"
    GENETIC_OPERATORS = "GENETIC OPERATORS"
    FLOOD_FILL = "FLOOD FILL"
    CAVITY_FILE = "cavity_file"
    DATA_FILES = "DATA FILES"
    LIGAND_DATA_FILE = "ligand_data_file"
    FLAGS = "FLAGS"
    TERMINATION = "TERMINATION"
    CONSTRAINTS = "CONSTRAINTS"
    COVALENT_BONDING = "COVALENT BONDING"
    SAVE_OPTIONS = "SAVE OPTIONS"
    FITNESS_FUNCTION_SETTINGS = "FITNESS FUNCTION SETTINGS"
    GOLD_FITFUNC_PATH = "gold_fitfunc_path"
    PROTEIN_DATA = "PROTEIN DATA"
    PROTEIN_DATAFILE = "protein_datafile"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepModelBuilderEnum:
    # configuration fields
    DATA = "data"
    DATA_INPUT_COLUMN = "input_column"
    DATA_RESPONSE_COLUMN = "response_column"
    DATA_TRAININGSET_FILE = "training_dataset_file"
    DATA_TESTSET_FILE = "test_dataset_file"

    # fixed tempfile names
    TMP_INPUT_CONFIG = "input_config.json"
    TMP_INPUT_DATA = "input_data.csv"
    TMP_OUTPUT_BEST_MODEL = "best_model.pkl"
    TMP_OUTPUT_BEST_PARAMETERS = "best_parameters.json"
    TMP_OUTPUT_PRODUCTION_MODEL = "production_model.pkl"

    # fields
    FIELD_KEY_PRODUCTION_MODEL = "production_model"
    FIELD_KEY_BEST_CONFIGURATION = "best_configuration"
    FIELD_KEY_INPUT_DATA = "input_data"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class TokenGuardEnum:
    TG = "token_guard"
    TG_PREFIX_EXECUTION = "prefix_execution"
    TG_BINARY_LOCATION = "binary_location"
    TG_TOKEN_POOLS = "token_pools"
    TG_WAIT_INTERVAL_SECONDS = "wait_interval_seconds"
    TG_WAIT_LIMIT_SECONDS = "wait_limit_seconds"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepPrepwizEnum:

    REMOVE_RES = "remove_res"
    LIGANDS = "ligands"
    COFACTOR_IDS = [
        "TDP",
        "FAD",
        "FMN",
        "NAD",
        "PNS",
        "COA",
        "PLP",
        "GSH",
        "BTN",
        "FFO",
        "B12",
        "ASC",
        "MQ7",
        "UQ1",
        "MGD",
        "H4B",
        "MDO",
        "SAM",
        "F43",
        "COM",
        "TP7",
        "HEA",
        "DPM",
        "PQQ",
        "TPQ",
        "TRQ",
        "LPA",
        "HEM",
    ]


class StepLigprepEnum:
    FILTER_FILE = "filter_file"

    # the SDF tag with <identifier>-# (where # is the number of the enumeration starting with '1')
    LIGPREP_VARIANTS = "s_lp_Variant"
    # number from 0 to 1 (sums up to 1 over all variants)
    LIGPREP_TAUTOMER_PROBABILITY = "r_lp_tautomer_probability"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepGlideEnum:
    # input specification parameters
    CONFIGURATION = "configuration"
    TIME_LIMIT_PER_TASK = "time_limit_per_task"
    MAESTRO_IN_FILE = "maestro_in_file"
    MAESTRO_IN_FILE_PATH = "path"

    # glide: fixed strings
    # the docking score (including "Epik" corrections")
    GLIDE_DOCKING_SCORE = "r_i_docking_score"
    # the "docking score" without "Epik" corrections
    GLIDE_GSCORE = "r_i_glide_gscore"
    # the index of the ligand in the input file (starting with '1')
    GLIDE_SOURCE_FILE_INDEX = "i_m_source_file_index"

    GLIDE_POSEVIEWER_FILE_KEY = "structures_pv.maegz"
    GLIDE_MAEGZ_DEFAULT_EXTENSION = "_pv.maegz"
    GLIDE_SDF_DEFAULT_EXTENSION = "_lib.sdfgz"
    GLIDE_LOG = ".log"
    GLIDE_SDF = ".sdf"

    FILL_DUMMY_CONFS = "fill_dummy_confs"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepMacromodelEnum:
    # COM file
    COM_FILE = "com_file"
    COM_FILE_PATH = "com_file.com"
    COM_FILE_DEFAULT = """ MMOD       0      1      0      0     0.0000     0.0000     0.0000     0.0000
 DEBG      55      0      0      0     0.0000     0.0000     0.0000     0.0000
 FFLD      16      1      0      0     1.0000     0.0000     0.0000     0.0000
 SOLV       3      1      0      0     0.0000     0.0000     0.0000     0.0000
 EXNB       0      0      0      0     0.0000     0.0000     0.0000     0.0000
 BDCO       0      0      0      0    89.4427 99999.0000     0.0000     0.0000
 READ       0      0      0      0     0.0000     0.0000     0.0000     0.0000
 CRMS       0      0      0      0     0.0000     0.8000     0.0000     0.0000
 LMCS    1000      0      0      0     0.0000     0.0000     3.0000     6.0000
 NANT       0      0      0      0     0.0000     0.0000     0.0000     0.0000
 MCNV       1      5      0      0     0.0000     0.0000     0.0000     0.0000
 MCSS       2      0      0      0    27.0000     0.0000     0.0000     0.0000
 MCOP       1      0      0      0     0.5000     0.0000     0.0000     0.0000
 DEMX       0    833      0      0    27.0000    54.0000     0.0000     0.0000
 MSYM       0      0      0      0     0.0000     0.0000     0.0000     0.0000
 AUOP       0      0      0      0   100.0000     0.0000     0.0000     0.0000
 AUTO       0      2      1      1     0.0000     1.0000     0.0000     2.0000
 CONV       2      0      0      0     0.0500     0.0000     0.0000     0.0000
 MINI       1      0   2500      0     0.0000     0.0000     0.0000     0.0000"""

    # fixed file names
    MAE_INPUT = "input_mol.mae"
    MAE_OUTPUT = "output_mol.mae"
    SDF_OUTPUT = "output_mol.sdf"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepPrimeEnum:

    RECEPTOR = "receptor"  # path to the receptor MAE file
    POSEVIEWER = "poseviewer"

    # fixed file names
    SDF_INPUT = "input_mol.sdf"
    MAE_INPUT = "input_mol.mae"
    MAE_COMPLEX = "complex.mae"
    MAE_OUTPUT = "complex-out.maegz"
    MMGBSA_SCORE = "r_psp_MMGBSA_dG_Bind"
    SDF_OUTPUT = "output_mol.sdf"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepTurbomoleEnum:
    EXECUTION_MODE = "execution_mode"
    SUCCESS = "success"
    FAILED = "failed"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepPantherEnum:
    # settings.additional
    PANTHER_LOCATION = "panther_location"
    PANTHER_CONFIG_FILE = "panther_config_file"
    OUTPUT_FILE = "output_file"
    PANTHER_CONFIG_DIR = "panther_config_file"
    FIELDS = "fields"

    # fields
    FIELD_KEY_NEGATIVE_IMAGE = "negative_image"
    FIELD_KEY_COORDINATES = "5-Center"
    FIELD_KEY_PDB_FILE = "1-Pdb file"

    # parameters
    FIELDS_PARAMETERS_LIB = {
        "2-Radius": "rad.lib",
        "3-Angle": "angles.lib",
        "4-Charge": "charges.lib",
    }

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepKallistoEnum:

    FEATURES = "features"
    SUCCESS = "success"
    FAILURE = "failure"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepJazzyEnum:
    SUCCESS = "success"
    FAILURE = "failure"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepShaepEnum:
    # field keys for storing data
    FIELD_KEY_NEGATIVE_IMAGE = "negative_image"
    NEG_IMAGE_EXT = "mol2"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepGromacsEnum:
    FIELDS = "fields"
    COFACTORS = "cofactors"
    FORCEFIELD = "forcefield"
    INPUT_FILE = "input_file"

    FIELD_KEY_STRUCTURE = "gro"
    FIELD_KEY_PDB = "pdb"
    FIELD_KEY_TOPOL = "top"
    FIELD_KEY_TPR = "tpr"
    FIELD_KEY_MDP = "mdp"
    FIELD_KEY_XTC = "xtc"
    FIELD_KEY_ITP = "itp"
    FIELD_KEY_LOG = "log"
    FIELD_KEY_EDR = "edr"
    FIELD_KEY_CPT = "cpt"
    FIELD_KEY_NDX = "ndx"
    PROPS = "props"
    FILE_SIZE_THRESHOLD = 2000000000

    MAKE_NDX_COMMAND = "make_ndx_command"
    INDEX_FLAG = "-n"

    #  magic strings associated with ligand parametrisation step
    FORCEFIELD_ITP = "forcefield.itp"
    LIGAND_ITP = "Ligand.itp"
    INCLUDE_LIG_ITP = '#include "Ligand.itp"'
    LIG_MOLECULE_GRP = "Ligand   1\n"
    COMPLEX_TOP = "Complex.top"
    COMPLEX_PDB = "Complex.pdb"
    PROTEIN_PDB = "Protein.pdb"
    PROTEIN_TOP = "Protein.top"
    LIGAND_PDB = "Ligand.pdb"
    LIGAND_MOL2 = "Ligand.mol2"
    STD_INDEX = "index.ndx"
    STD_TOPOL = "topol.top"
    STD_EDR = "ener.edr"
    STD_CPT = "state.cpt"
    STD_TPR = "topol.tpr"
    STD_LOG = "md.log"
    STD_XTC = "traj.xtc"
    STD_TRR = "traj.trr"
    STD_STRUCTURE = "confout.gro"
    POSRE_LIG = "posre_lig.itp"
    CHARGE_METHOD = "charge_method"
    FORCE_CONSTANTS = "1000 1000 1000"
    LIG_ID = "lig_id"
    COUPLING_GROUP = "Other"
    MMPBSA_IN = "mmpbsa.in"
    THREADS = "threads"
    GROMACS_LOAD = "module load GROMACS/2021-fosscuda-2019a-PLUMED-2.7.1-Python-3.7.2"
    AMBERTOOLS_PREFIX = "ambertools_prefix"
    WATER_AND_IONS = "Water_and_ions"
    PROTEIN_OTHER = "Protein_Other"
    SIM_COMPLETE = "Finished mdrun"
    AUTO = "auto"
    TC_GRPS = "tc-grps"
    CLUSTERS_NUMBER = "clustersNumber"
    LENGTHS = "lengths"
    COUPLING_GROUPS = "coupling_groups"
    DEFAULT_MMPBSA_IN = "src/icolos/config/amber/default_mmpbsa.in"
    PARAM_METHOD = "param_method"
    MMGBSA_DG = "MMGBSA_DG"
    GAFF = "gaff"
    OPENFF = "openff"
    RESTRAINTS = "restraints"
    WATER_POSRE = """
#ifdef POSRES_WATER
[ position_restraints ]
;  i funct       fcx        fcy        fcz
   1    1       1000       1000       1000
#endif\n"""

    MULTIDIR = "multidir"
    REPLICAS = "replicas"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepOpenFFEnum:
    UNIQUE_MOLS = "unique_molecules"
    METHOD = "method"
    PARMED = "parmed"
    INTERCHANGE = "interchange"
    FORCEFIELD = "off_forcefield"


class StepCavExploreEnum:
    FIELD_KEY_DTR = "dtr"
    FIELD_KEY_CMS = "cms"
    FIELD_KEY_DX = "dx"

    # settings.additional
    CAVITY_LOCATION = "cavity_location"
    CAVITY_CONFIG_FILE = "cavity_config_file"
    OUTPUT_FILE = "output_file"
    CAVITY_CONFIG_DIR = "cavity_config_dir"
    FIELDS = "fields"
    SELECTION_TEXT = "selection_text"
    PROTEIN = "protein"
    NAME_CA = "name CA"
    FRAME_LIST_FILE = "list_of_frames.txt"
    MDPOCKET_COMMAND = "mdpocket"
    MPI_THREADS = "mpi_threads"
    EPS = "eps"
    MIN_SAMPLES = "min_samples"
    ISO_VALUE = "iso_value"
    TRAJ_TYPE = "format"
    THRESHOLD = "threshold"

    # add own fixed strings and import in step
    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepFepPlusEnum:
    FIELD_KEY_POSEVIEWER = "poseviewer"
    RECEPTOR_MAEGZ = "receptor.maegz"
    STRUCT_SPLIT_BASE = "split"
    STRUCTCAT_MAEGZ_OUTFILE = "concatenated.mae"
    STRUCTCAT_SDF_OUTFILE = "concatenated.sdf"
    FEP_MAPPER_OUTPUT = "out"
    FMP_OUTPUT_FILE = "out.fmp"
    LOGFILE = "multisim.log"
    EDGE_HEADER_LINE = "* Edge calculated properties (units in kcal/mol)"
    NODE_HEADER_LINE = "* Node calculated properties (units in kcal/mol)"
    SIMULATION_PROTOCOL = "* Simulation Protocol"
    SIMILARITY = "* Similarity"
    DATA_TERMINUS = "fep_mapper_cleanup: Loading output graph"
    HOST_FLAG = "-HOST"
    WAIT_FLAG = "-WAIT"
    JOBNAME_FLAG = "-JOBNAME"
    REFERENCE_DG = "ref_dg"
    JOBID_STRING = "JobId:"
    XRAY_STRUCTURES = "xray_structures"
    XRAY_SPLIT = "xray_split"
    RETRIES = "-RETRIES"

    FILE_NAME = "--name"
    FEP_EXEC_COMPLETE = "Multisim completed."
    FEP_EXEC_PARTIAL_COMPLETE = "Multisim partially completed."

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

        # prohibit any attempt to set any values

    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepFilterEnum:
    FILTER_LEVEL = "filter_level"
    CONFORMERS = "conformers"
    COMPOUNDS = "compounds"
    HIGHEST_IS_BEST = "highest_is_best"
    ENUMERATIONS = "enumerations"
    CRITERIA = "criteria"
    AGGREGATION = "aggregation"
    RETURN_N = "return_n"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

        # prohibit any attempt to set any values

    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepDesmondEnum:

    PREPROCESS_MSJ = "config.msj"
    PRODUCTION_MSJ = "production.msj"
    PRODUCTION_CFG = "prod.cfg"
    MSJ_FIELDS = "msj_fields"
    CFG_FIELDS = "cfg_fields"
    SETUP_MSJ_FIELDS = "setup_msj_fields"
    CONFIG = "config"
    TOKEN_STR = "DESMOND_GPGPU:16"

    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

        # prohibit any attempt to set any values

    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepPdbFixerEnum:
    ADD_MISSING_HYDROGENS = "add_missing_hydrogens"
    ADD_MISSING_ATOMS = "add_missing_atoms"
    FIND_MISSING_ATOMS = "find_missing_atoms"
    FIND_MISSING_RESIDUES = "find_missing_residues"
    REPLACE_NONSTANDARD_RESIDUES = "replace_nonstandard_residues"
    REMOVE_CHAINS = "remove_chains"


class StepDSSPEnum:
    pass


class StepCressetEnum:
    SUCCESS = "success"


class StepAutoDockVinaEnum:

    ADV_RECEPTOR_PATH = "receptor_path"
    ADV_SEED = "seed"
    ADV_SEARCH_SPACE = "search_space"
    ADV_SEARCH_SPACE_CENTER_X = "--center_x"
    ADV_SEARCH_SPACE_CENTER_Y = "--center_y"
    ADV_SEARCH_SPACE_CENTER_Z = "--center_z"
    ADV_SEARCH_SPACE_SIZE_X = "--size_x"
    ADV_SEARCH_SPACE_SIZE_Y = "--size_y"
    ADV_SEARCH_SPACE_SIZE_Z = "--size_z"

    CONFIGURATION = "configuration"
    NUMBER_POSES = "number_poses"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepAutoDockVinaTargetPreparationEnum:

    ADV_PDBQT = ".pdbqt"
    INPUT_RECEPTOR_PDB = "input_receptor_pdb"
    OUTPUT_RECEPTOR_PDBQT = "output_receptor_pdbqt"
    PH = "pH"
    EXTRACT_BOX = "extract_box"
    EXTRACT_BOX_REFERENCE_LIGAND_PATH = "reference_ligand_path"
    EXTRACT_BOX_REFERENCE_LIGAND_FORMAT = "reference_ligand_format"
    EXTRACT_BOX_REFERENCE_LIGAND_FORMAT_PDB = "PDB"
    EXTRACT_BOX_REFERENCE_LIGAND_FORMAT_SDF = "SDF"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepGoldTargetPreparationEnum:

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class StepActiveLearningEnum:

    ORACLE_TYPE = "oracle_type"
    ORACLE_CONFIG = "oracle_config"
    SMILES = "SMILES"
    MOLECULE = "Molecule"
    VIRTUAL_LIB = "virtual_lib"
    FRAGMENTS = "fragments"
    INIT_SAMPLE_FACTOR = "init_sample_factor"
    MORGAN_FP = "MorganFP"
    SOAP_VECTOR = "soap_vector"
    ID = "ID"
    IDX = "IDX"
    GRAPH = "graph"
    FRACTION_PER_EPOCH = "fraction_per_epoch"
    MAX_SAMPLED_FRACTION = "max_sampled_fraction"
    WARMUP = "warmup"
    N_ROUNDS = "n_rounds"
    BATCH_SIZE = "batch_size"
    ACQUISITION_FUNCTION = "acquisition_function"
    MODEL = "model"
    EVALUATE = "evaluate"
    CRITERIA = "criteria"
    VALIDATION_LIB = "validation_lib"
    DYNAMIC_STOP = "dynamic_stop"
    REPLICAS = "replicas"
    EPSILON = "epsilon"
    LATENT_DISTANCE = "latent_distance"
    GREEDY = "greedy"
    EI = "ei"

    # Prospective Active Learning in REINVENT
    # -----------------------------------------------
    # General Parameters
    EPOCHS = "number_of_reinvent_epochs"
    WARMUP = "warmup"
    RETRAIN = "retrain"
    INITIAL_POOLING_EPOCHS = "initial_pooling_epochs"
    ACQUISITION_BATCH_SIZE = "acquisition_batch_size"
    ORACLE_CONFIG = "oracle_config"
    ORACLE_LABEL = "oracle_label"
    SAVE_DIR = "save_dir"

    # Surrogate Models
    SURROGATE_MODEL_TYPE = "surrogate_model_type"
    RANDOM_FOREST_REGRESSOR = "rf"
    SUPPORT_VECTOR_REGRESSOR = "svr"

    # Acquisition Functions
    RANDOM = "random"
    GREEDY = "greedy"
    GREEDY_EPSILON = "greedy_epsilon"
    PI = "pi"  # probability of improvement
    EI = "ei"  # expected improvement
    TS = "ts"  # thompson sampling
    UCB = "ucb"  # upper confidence bound
    TANIMOTO = "tanimoto"
    TANIMOTO_TRUE_BELIEVER = "tanimoto_true_believer"



from pydantic import BaseModel

from icolos.utils.general.icolos_exceptions import StepFailed
from icolos.core.workflow_steps.io.base import StepIOBase
from icolos.core.workflow_steps.step import _LE


class StepInitializeCompound(StepIOBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    def execute(self):
        if len(self.data.compounds) == 0:
            raise StepFailed(
                "Compound initialization step failed - no Compound objects generated."
            )
        self._logger.log(
            f"Step {self.get_step_id()} initialized {len(self.get_compounds())} compounds.",
            _LE.INFO,
        )



from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
from icolos.utils.execute_external.execute import Executor
from pydantic import BaseModel
from copy import deepcopy
from icolos.core.workflow_steps.step import _LE
import os
import pandas as pd


class StepResidueScanning(StepSchrodingerBase, BaseModel):
    """
    Interface to Schrodinger's PrepWizard program for protein prep
    """

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(executor=Executor)
        self._check_backend_availability()

    def _parse_args(self):
        parameters = deepcopy(self.settings.arguments.parameters)
        arguments = []
        if len(self.settings.arguments.flags) > 0:
            for flag in self.settings.arguments.flags:
                arguments.append(str(flag))
        if parameters:
            for key in parameters.keys():
                arguments.append(key)
                if parameters[key] is not None and parameters[key] != "":
                    arguments.append(str(parameters[key]))
        input_file = self.data.generic.get_argument_by_extension("maegz")
        arguments.append(input_file)
        print(arguments)
        return arguments

    def _parse_output(self, tmp_dir: str):
        # run the conversion script, generate csv for the mutants, attach to compounds
        out_file = [f for f in os.listdir(tmp_dir) if f.endswith("out.maegz")][0]
        command = f"$SCHRODINGER/utilities/proplister {out_file} -p s_bioluminate_Mutations -p r_bioluminate_delta_Stability -p r_bioluminate_delta_Affinity -c -o props_out.csv"
        self._backend_executor.execute(
            command, arguments=[], check=True, location=tmp_dir
        )

        # parse csv file, attach to compounds
        compounds = self.get_compounds()
        props_df = pd.read_csv("props_out.csv")
        for comp in compounds:
            mol = comp.get_enumerations()[0].get_molecule()

            # find the row by the name
            mol_name = mol.GetProp("_Name")
            row = props_df.loc[
                props_df["s_bioluminate_Mutations"].str.contains(mol_name)
            ]
            try:
                mol.SetProp(
                    "r_bioluminate_delta_Stability",
                    str(float(row["r_bioluminate_delta_Stability"])),
                )
                mol.SetProp(
                    "r_bioluminate_delta_Affinity",
                    str(float(row["r_bioluminate_delta_Affinity"])),
                )
                self._logger.log(
                    f"Parsed properties for mol {mol_name}: {mol.GetProp('r_bioluminate_delta_Affinity')}, {mol.GetProp('r_bioluminate_delta_Stability')}",
                    _LE.DEBUG,
                )
            except (KeyError, TypeError):
                self._logger.log("Failed to parse results df", _LE.WARNING)
                mol.SetProp(
                    "r_bioluminate_delta_Stability",
                    str(0.00),
                )
                mol.SetProp(
                    "r_bioluminate_delta_Affinity",
                    str(0.00),
                )

    def execute(self):
        tmp_dir = self._make_tmpdir()
        print(tmp_dir)
        args = self._parse_args()
        self.data.generic.write_out_all_files(tmp_dir)

        command = f"$SCHRODINGER/run residue_scanning_backend.py"
        self._logger.log("executing residue scanning", _LE.DEBUG)
        result = self._backend_executor.execute(
            command=command, arguments=args, check=True, location=tmp_dir
        )
        self._parse_output(tmp_dir)

        # self._remove_temporary(tmp_dir)



import os
import shutil
import tempfile
from typing import List, Tuple

from pydantic import BaseModel, Field
from rdkit import Chem
from copy import deepcopy

from icolos.utils.enums.step_enums import StepAutoDockVinaEnum, StepBaseEnum
from icolos.utils.execute_external.autodockvina import AutoDockVinaExecutor
from icolos.utils.execute_external.openbabel import OpenBabelExecutor
from icolos.utils.general.icolos_exceptions import StepFailed
from icolos.utils.general.files_paths import gen_tmp_file
from icolos.core.containers.compound import Conformer
from icolos.utils.enums.program_parameters import AutoDockVinaEnum, OpenBabelEnum
from icolos.core.workflow_steps.step import _LE, StepBase
from icolos.utils.general.parallelization import Subtask, SubtaskContainer, Parallelizer

_SBE = StepBaseEnum
_ADE = AutoDockVinaEnum()
_OBE = OpenBabelEnum()
_SAE = StepAutoDockVinaEnum()


class ADVSearchSpace(BaseModel):
    center_x: float = Field(alias="--center_x", default=None)
    center_y: float = Field(alias="--center_y", default=None)
    center_z: float = Field(alias="--center_z", default=None)
    size_x: float = Field(alias="--size_x", default=15.0)
    size_y: float = Field(alias="--size_y", default=15.0)
    size_z: float = Field(alias="--size_z", default=15.0)


class ADVConfiguration(BaseModel):
    seed: int = 42
    number_poses: int = 1
    search_space: ADVSearchSpace = ADVSearchSpace()
    receptor_path: str = None


class ADVAdditional(BaseModel):
    configuration: ADVConfiguration = ADVConfiguration()
    grid_ids: List[str] = ["grid0"]


class StepAutoDockVina(StepBase, BaseModel):

    _openbabel_executor: OpenBabelExecutor = None
    adv_additional: ADVAdditional = None

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=AutoDockVinaExecutor)
        self._check_backend_availability()

        # initialize the executor for all "OpenBabel"
        self._openbabel_executor = OpenBabelExecutor()
        if not self._openbabel_executor.is_available():
            raise StepFailed(
                "AutoDock Vina requires OpenBabel execution, initialization failed."
            )

        # set ADV specific settings and ensure that each molecule gets its own sublist
        self.adv_additional = ADVAdditional(**self.settings.additional)
        self.execution.parallelization.max_length_sublists = 1

    def _set_docking_score(self, conformer: Chem.Mol) -> bool:
        try:
            result_tag_lines = conformer.GetProp(_ADE.REMARK_TAG).split("\n")
            result_line = [
                line for line in result_tag_lines if _ADE.RESULT_LINE_IDENTIFIER in line
            ][0]
            parts = result_line.split()
            docking_score = parts[_ADE.RESULT_LINE_POS_SCORE]
        except KeyError:
            return False
        conformer.SetProp(_SBE.ANNOTATION_TAG_DOCKING_SCORE, str(docking_score))
        return True

    def _write_molecule_to_pdbqt(self, path: str, molecule: Chem.Mol) -> bool:
        # generate temporary copy as PDB
        _, tmp_pdb = gen_tmp_file(suffix=".pdb", dir=os.path.dirname(path))
        Chem.MolToPDBFile(molecule, filename=tmp_pdb)

        # translate the pdb into a pdbqt including partial charges
        # Note: In contrast to the target preparation,
        # we will use a tree-based flexibility treatment here -
        # thus, the option "-xr" is NOT used.
        arguments = [
            tmp_pdb,
            _OBE.OBABEL_OUTPUT_FORMAT_PDBQT,
            "".join([_OBE.OBABEL_O, path]),
            _OBE.OBABEL_PARTIALCHARGE,
            _OBE.OBABEL_PARTIALCHARGE_GASTEIGER,
        ]
        self._openbabel_executor.execute(
            command=_OBE.OBABEL, arguments=arguments, check=False
        )

        if os.path.exists(path):
            return True
        else:
            return False

    def _generate_temporary_input_output_files(
        self, batch: List[List[Subtask]]
    ) -> Tuple[List[str], List[str], List[str], List[str]]:
        tmp_output_dirs = []
        tmp_input_paths = []
        tmp_output_paths = []
        enumeration_ids = []

        for next_subtask_list in batch:
            # for "AutoDock Vina", only single molecules can be handled so every sublist is
            # guaranteed at this stage to have only one element
            if len(next_subtask_list) > 1:
                self._logger.log(
                    f"Subtask list length for ADV is > 1 ({len(next_subtask_list)}), only the first element will be processed.",
                    _LE.WARNING,
                )
            subtask = next_subtask_list[0]

            # generate temporary input files and output directory
            cur_tmp_output_dir = tempfile.mkdtemp()
            _, cur_tmp_input_pdbqt = gen_tmp_file(
                suffix=".pdbqt", dir=cur_tmp_output_dir
            )
            _, cur_tmp_output_sdf = gen_tmp_file(suffix=".sdf", dir=cur_tmp_output_dir)

            # try to write the enumeration molecules out as PDBQT files
            enumeration = subtask.data
            mol = deepcopy(enumeration.get_molecule())
            if mol is None:
                shutil.rmtree(cur_tmp_output_dir)
                self._logger.log(
                    f"Enumeration {enumeration.get_index_string()} did not hold a valid RDkit molecule - skipped.",
                    _LE.DEBUG,
                )
                continue
            if not self._write_molecule_to_pdbqt(cur_tmp_input_pdbqt, mol):
                self._logger.log(
                    f"Could not generate PDBQT intermediate file from enumeration {enumeration.get_index_string()} - skipped.",
                    _LE.DEBUG,
                )
                continue

            # also store all the paths in case it succeeded -> these will be used later, failures will be ignored
            tmp_output_dirs.append(cur_tmp_output_dir)
            tmp_input_paths.append(cur_tmp_input_pdbqt)
            tmp_output_paths.append(cur_tmp_output_sdf)
            enumeration_ids.append(enumeration.get_index_string())

        return tmp_output_dirs, tmp_input_paths, tmp_output_paths, enumeration_ids

    def _execute_autodockvina(self):
        # get number of sublists in batch and initialize Parallelizer
        adv_parallelizer = Parallelizer(func=self._run_subjob)

        # continue until everything is successfully done or number of retries have been exceeded
        while self._subtask_container.done() is False:
            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())

            # generate paths and initialize molecules (so that if they fail, this can be covered)
            (
                tmp_output_dirs,
                tmp_input_paths,
                tmp_output_paths,
                enumeration_ids,
            ) = self._generate_temporary_input_output_files(next_batch)

            # execute the current batch in parallel; hand over lists of parameters (will be handled by Parallelizer)
            # also increment the tries and set the status to "failed" (don't do that inside subprocess, as data is
            # copied, not shared!)
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]
            adv_parallelizer.execute_parallel(
                input_path_pdbqt=tmp_input_paths, output_path_sdf=tmp_output_paths
            )

            # parse the output of that particular batch and remove temporary files
            self._parse_adv_output_batch(
                tmp_input_paths=tmp_input_paths,
                tmp_output_paths=tmp_output_paths,
                enumeration_ids=enumeration_ids,
                next_batch=next_batch,
            )

            # clean-up
            self._remove_temporary(tmp_output_dirs)

            # print the progress for this execution
            self._log_execution_progress()

    def _parse_adv_output_batch(
        self,
        tmp_input_paths: List[str],
        tmp_output_paths: List[str],
        enumeration_ids: List[str],
        next_batch: List[List[Subtask]],
    ):

        for i in range(len(next_batch)):
            subtask = next_batch[i][0]
            tmp_output_path = tmp_output_paths[i]
            enumeration_id = enumeration_ids[i]
            grid_id = self.adv_additional.grid_ids[0]
            grid_path = self.adv_additional.configuration.receptor_path

            # this is a protection against the case where empty (file size == 0 bytes) files are generated due to
            # a failure during docking
            if (
                not os.path.isfile(tmp_output_path)
                or os.path.getsize(tmp_output_path) == 0
            ):
                continue

            mol_supplier = Chem.SDMolSupplier(tmp_output_path, removeHs=False)
            for mol in mol_supplier:
                if mol is None:
                    continue

                # add the information on the actual grid used
                mol.SetProp(_SBE.ANNOTATION_GRID_ID, str(grid_id))
                mol.SetProp(_SBE.ANNOTATION_GRID_PATH, str(grid_path))
                mol.SetProp(_SBE.ANNOTATION_GRID_FILENAME, os.path.basename(grid_path))

                # if no docking score is attached (i.e. the molecule is a receptor or so, skip it)
                if self._set_docking_score(mol) is not True:
                    continue

                # add molecule to the appropriate ligand
                for compound in self.get_compounds():
                    for enumeration in compound:
                        if enumeration.get_index_string() == enumeration_id:
                            new_conformer = Conformer(
                                conformer=mol,
                                conformer_id=None,
                                enumeration_object=enumeration,
                            )
                            enumeration.add_conformer(new_conformer, auto_update=True)
                            subtask.set_status_success()
                            break

    def _delay_file_system(self, path) -> bool:
        return self._wait_until_file_generation(
            path=path, interval_sec=2, maximum_sec=10
        )

    def _run_subjob(self, input_path_pdbqt: str, output_path_sdf: str):

        config = self.adv_additional.configuration

        # set up arguments list and execute
        _, tmp_pdbqt_docked = gen_tmp_file(
            suffix=".pdbqt", dir=os.path.dirname(input_path_pdbqt)
        )
        arguments = [
            _ADE.VINA_RECEPTOR,
            config.receptor_path,
            _ADE.VINA_LIGAND,
            input_path_pdbqt,
            _ADE.VINA_CPU,
            str(1),
            _ADE.VINA_SEED,
            config.seed,
            _ADE.VINA_OUT,
            tmp_pdbqt_docked,
            _ADE.VINA_CENTER_X,
            str(config.search_space.center_x),
            _ADE.VINA_CENTER_Y,
            str(config.search_space.center_y),
            _ADE.VINA_CENTER_Z,
            str(config.search_space.center_z),
            _ADE.VINA_SIZE_X,
            str(config.search_space.size_x),
            _ADE.VINA_SIZE_Y,
            str(config.search_space.size_y),
            _ADE.VINA_SIZE_Z,
            str(config.search_space.size_z),
            _ADE.VINA_NUM_MODES,
            config.number_poses,
        ]

        self._backend_executor.execute(
            command=_ADE.VINA, arguments=arguments, check=True
        )
        self._delay_file_system(path=tmp_pdbqt_docked)

        # translate the parsed output PDBQT into an SDF, via regular pdb to retain hydrogens!
        # This appears to be a bug in obabel 3.1
        # TODO: move to a direct conversino once the obabel bug is fixed
        arguments = [
            tmp_pdbqt_docked,
            _OBE.OBABEL_INPUTFORMAT_PDBQT,
            _OBE.OBABEL_OUTPUT_FORMAT_PDB,
            # clip the extension to pdb
            "".join([_OBE.OBABEL_O, tmp_pdbqt_docked[:-2]]),
        ]
        self._openbabel_executor.execute(
            command=_OBE.OBABEL, arguments=arguments, check=True
        )
        # convert resulting pdb to sdf
        arguments = [
            _OBE.OBABEL_INPUTFORMAT_PDB,
            tmp_pdbqt_docked[:-2],
            _OBE.OBABEL_OUTPUT_FORMAT_SDF,
            # clip the extension to pdb
            "".join([_OBE.OBABEL_O, output_path_sdf]),
            "-h",
        ]
        self._openbabel_executor.execute(
            command=_OBE.OBABEL, arguments=arguments, check=True
        )
        self._delay_file_system(path=output_path_sdf)

    def execute(self):
        # Note: This step only supports one grid at a time, ensemble docking is taken care of at the workflow level!

        # in order to be able to efficiently execute ADV on the enumeration level, all of them have to be unrolled
        # Note: As they retain their respective Compound object, the attribution later on is simple
        all_enumerations = []
        for compound in self.get_compounds():
            all_enumerations = all_enumerations + compound.get_enumerations()
            for enumeration in compound:
                enumeration.clear_conformers()

        # split into sublists, according to the settings
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_enumerations)

        # execute ADV
        self._execute_autodockvina()



from icolos.utils.general.parallelization import Parallelizer, SubtaskContainer
from pydantic import BaseModel
from icolos.core.workflow_steps.fpocket.base import StepFpocketBase
from icolos.utils.enums.step_enums import StepCavExploreEnum
from icolos.utils.execute_external.execute import Executor
from icolos.core.workflow_steps.step import _LE
from sklearn.cluster import DBSCAN
from collections import Counter
import numpy as np
import re
import os

_SFP = StepCavExploreEnum()


class StepMDpocket(StepFpocketBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # self._initialize_backend(executor=MPIExecutor)
        self._initialize_backend(executor=Executor)

        # set max_length_sublists to 1
        self.execution.parallelization.max_length_sublists = 1

    def _create_density_grid_file(self, tmp_dir: str, iso_value: float):
        """creates a density grid from the .dx-file into a .pdb-file, heavily influenced by extractISOPdb.py provided
        by fpocket"""
        pass

    def _cluster_pockets(self, tmp_dir, eps, min_samples, threshold, iso_value):
        """
        Clusters points from the initial MDpocket density grid, at a certain iso value
        """
        pass

    def _save_pocket_files(self, tmp_dir, data, labels):
        """saves the individual pockets as individual pdbs to be used with mdpocket"""
        pass

    def _run_mdpocket_selected_pocket(self, tmp_dir):
        """runs the second mdpocket command for fpocket3"""
        pass

    def _prepare_batch_inputs(self, batch, tmp_dir):
        tmp_dirs = []
        args = []
        for next_subtask_list in batch:
            tmp_dirs.append(tmp_dir)
            for (
                subtask
            ) in (
                next_subtask_list
            ):  # enforced only one task per subtask, otherwise it makes no sense
                args.append(subtask.data)  # append the arguments list
        return tmp_dirs, args

    def _execute_mdpocket(self, tmp_dir, arguments):

        self._backend_executor.execute(
            command=_SFP.MDPOCKET_COMMAND,
            arguments=arguments,
            location=tmp_dir,
            check=True,
        )

    def execute(self):
        pass
        



import os
from typing import Tuple

from icolos.loggers.base_logger import BaseLogger

from icolos.utils.enums.composite_agents_enums import WorkflowEnum
from icolos.utils.enums.entry_points import ExecutorEnum
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.general.files_paths import move_up_directory

_WE = WorkflowEnum()
_LE = LoggingConfigEnum()
_EE = ExecutorEnum()


def parse_global(g_input, logger: BaseLogger) -> dict:
    if g_input is not None:
        if not isinstance(g_input, list):
            g_input = [g_input]
        g_vars = {}
        for new_var in g_input:
            parts = new_var.split(":")
            if len(parts) != 2:
                logger.log(
                    f"Ignoring global input {new_var} set by command-line, as they must have one key and one value, separated by ':'.",
                    _LE.WARNING,
                )
                continue
            g_vars[parts[0]] = parts[1]
            logger.log(
                f'Parsed global input "{parts[0]}" (value: "{parts[1]}").', _LE.DEBUG
            )
        return g_vars
    else:
        return {}


def add_global(configuration: dict, g_vars: dict, field: str) -> dict:
    """This function adds (and overwrites) values for global settings and variables. Parameter "field" selects,
    which key is to be used in the header region."""
    header = configuration[_WE.WORKFLOW][_WE.HEADER]
    if field not in header.keys():
        header[field] = {}
    for key, value in g_vars.items():
        header[field][key] = value
    return configuration


def get_runtime_global_variables(args_conf: str, entry_point_path: str) -> dict:
    return {  # current workdir
        _EE.RUNTIME_GLOBAL_VARIABLE_WORKDIR: os.getcwd(),
        # directory where the entry point lies
        _EE.RUNTIME_GLOBAL_VARIABLE_ENTRYPOINTDIR: os.path.dirname(entry_point_path),
        # directory where the JSON lies
        _EE.RUNTIME_GLOBAL_VARIABLE_CONFIGDIR: os.path.dirname(
            os.path.abspath(args_conf)
        ),
        # top level of Icolos icolos package
        _EE.RUNTIME_GLOBAL_VARIABLE_PACKAGEDIR: move_up_directory(
            os.path.dirname(entry_point_path), 3
        ),
    }


def parse_header(conf: dict, args, entry_point_path: str, logger: BaseLogger) -> dict:
    # parse global variables from command-line
    global_vars_CLI = parse_global(g_input=args.global_variables, logger=logger)
    conf = add_global(conf, global_vars_CLI, _WE.GLOBAL_VARIABLES)

    # add run-specified global variables (the current directory, the JSONs directory, ...)
    conf = add_global(
        conf,
        get_runtime_global_variables(args.conf, entry_point_path),
        _WE.GLOBAL_VARIABLES,
    )

    # update global settings; if they are not supported, pydantic will complain later on
    # TODO: at the moment this implementation ignores stuff that is not understood (e.g. when a typo occurs); this should fail
    global_settings_CLI = parse_global(g_input=args.global_settings, logger=logger)
    conf = add_global(conf, global_settings_CLI, _WE.GLOBAL_SETTINGS)
    return conf


def get_version_number() -> str:
    try:
        # this requires python >= 3.8
        from importlib import metadata

        return metadata.version("icolos")
    except ImportError:
        return None


def get_config_version_number(conf: dict) -> str:
    try:
        return str(conf[_WE.WORKFLOW][_WE.HEADER][_WE.VERSION])
    except KeyError:
        return None


def log_version_number(logger: BaseLogger):
    version = get_version_number()
    if version is None:
        logger.log(f"Could not obtain Icolos version.", _LE.WARNING)
        version = "unknown"
    logger.log(f"Icolos version {version} initialized.", _LE.INFO)


def version_match(conf: dict) -> bool:
    version_config = get_config_version_number(conf)
    version_installation = get_version_number()
    if (
        version_config is None
        or version_installation is None
        or version_config != version_installation
    ):
        return False
    return True


def get_versions_as_strings(conf: dict) -> Tuple[str, str]:
    version_config = get_config_version_number(conf)
    version_installation = get_version_number()
    version_config = "unknown" if version_config is None else version_config
    version_installation = (
        "unknown" if version_installation is None else version_installation
    )
    return version_config, version_installation



from typing import Dict, List, Union
from pydantic import BaseModel
from icolos.core.composite_agents.workflow import WorkFlow

from icolos.core.flow_control.flow_control import BaseStepConfig, FlowControlBase
from copy import deepcopy
from icolos.core.workflow_steps.step import _LE
from icolos.core.step_dispatch.dispatcher import StepDispatcher
from icolos.utils.enums.composite_agents_enums import WorkflowEnum
from icolos.utils.enums.step_enums import StepBaseEnum
from icolos.core.workflow_steps.step import StepBase
from icolos.utils.enums.step_enums import IteratorEnum
from icolos.core.composite_agents.workflow import WorkflowHeaderParameters
import os

_IE = IteratorEnum
_SBE = StepBaseEnum
_WE = WorkflowEnum()


class IterSettingsParameters(BaseModel):
    # unpacked version of StepSettingsParameters
    flags: List = []
    parameters: Dict = {}
    additional: Dict = {}
    work_dir: Union[List, str] = []


class IterParallelizer(BaseModel):

    # if true, steps must be totally independent, the iterator step
    parallelize: bool = False
    jobs: int = 1
    dependent_steps: int = None


class IterSettings(BaseModel):
    # dictionary of settings to change

    # settings: {step_id: {IterSettingsParameters}}
    settings: Dict[str, IterSettingsParameters] = {}
    iter_mode: _IE = _IE.N_ITERS
    n_iters: int = None
    remove_temporary_files: bool = True
    single_directory: bool = False
    parallelizer_settings: IterParallelizer = IterParallelizer()


class StepIterator(FlowControlBase, BaseModel):
    """
    Implements iterator mechanism:
    wraps one or multiple steps and generates n copies of that set of steps according to iter_config
    Becomes master job when parallize=True, using  Icolos JobControl to interface with external resources
    """

    # holds the dict of iterables for the bits to chang
    iter_settings: IterSettings = IterSettings()
    dispatcher: StepDispatcher = None

    def __init__(self, **data):
        super().__init__(**data)

        self.dispatcher = self._initialize_workflows()

    def _modify_settings(
        self, settings: BaseStepConfig, step_config: BaseStepConfig, i: int
    ):
        base_conf = deepcopy(step_config)
        iter_settings = deepcopy(settings)
        if settings.flags:
            settings.flags = {"flags": settings.flags}

            # iterate over the flags
            # if it's been converted, hence there are flags to be applied
            base_conf.settings.arguments.flags.append(iter_settings.flags.values()[i])
        for (
            key,
            val,
        ) in iter_settings.parameters.items():
            base_conf.settings.arguments.parameters[key] = val[i]
        for (
            key,
            val,
        ) in iter_settings.additional.items():
            # however many lists of n items
            base_conf.settings.additional[key] = val[i]

        return base_conf

    def _initialize_n_iters(self) -> List[WorkFlow]:
        """
        Initialise n identical copies of the same step config
        """
        workflows = []
        for i in range(self.iter_settings.n_iters):
            workflow_steps = []
            list_step_conf = deepcopy(self.base_config)

            # hand all steps over to the config updater
            step_confs = self._update_config(list_step_conf, f"run_{i}")
            for step in step_confs:
                workflow_steps.append(step.as_dict())
            wf_config = {
                _WE.HEADER: {
                    _WE.ID: f"workflow_{i}",
                    _WE.ENVIRONMENT: {_WE.ENVIRONMENT_EXPORT: []},
                    _WE.GLOBAL_VARIABLES: {},
                    _WE.GLOBAL_SETTINGS: {
                        "remove_temporary_files": self.iter_settings.remove_temporary_files,
                        "single_directory": self.iter_settings.single_directory,
                    },
                },
                _WE.STEPS: workflow_steps,
            }
            # manually set some things for these workflows

            workflows.append(WorkFlow(**wf_config))
        return workflows

    def _initialize_settings(self) -> List:
        """
        Iterate through all settings step-wise, changing all setting blocks simultaneously, returning n initialised steps for n
        """
        init_steps = []
        for i in range(self.iter_settings.n_iters):

            # iterate over the steps in the base config, and the corresponding settings, if these are to be modified
            step_sublist = []
            for step_config in deepcopy(self.base_config):

                # check if we need to iterate through settings in this step, else just use the base config
                if step_config.step_id in self.iter_settings.settings.keys():
                    settings = self.iter_settings.settings[step_config.step_id]
                    step_sublist.append(self._modify_settings(settings, step_config, i))
                else:
                    step_sublist.append(step_config)

            # update all configs with references to updated step_ids etc
            formatted_configs = self._update_config(step_sublist, f"run_{i}")
            for step_conf in formatted_configs:
                initialized_step = self._initialize_step_from_dict(step_conf._as_dict())
                init_steps.append(initialized_step)
        return init_steps

    def _initialize_workflows(self) -> StepDispatcher:
        """
        Handle step init according to config
        Returns a Step-like JobControl that holds a list of separate workflows, which will be executed according to the parallelization scheme.
        """
        workflows = []
        if self.iter_settings.iter_mode == _IE.N_ITERS:
            # simplest mode, generate n independent workflows with the same set of steps
            workflows += self._initialize_n_iters()

        elif self.iter_settings.iter_mode == _IE.SINGLE:
            raise NotImplementedError

        else:
            raise NotImplementedError

        self._logger.log(
            f"Initialized {len(workflows)} jobs for step  {self.base_config[0].step_id}",
            _LE.DEBUG,
        )
        dispatcher = StepDispatcher(
            step_id="step_dispatcher",
            type=_SBE.STEP_DISPATCHER,
            workflows=workflows,
            parallel_execution=self.iter_settings.parallelizer_settings,
        )
        return dispatcher

    def _update_config(
        self, step_conf: List[BaseStepConfig], run_id: str
    ) -> List[BaseStepConfig]:
        """
        Manages modifications to each step in the config:
        * writeout paths are updated to separate output from each of the runs
        """

        formatted_confs = []
        for conf in step_conf:

            # modify the writeout paths: add a key_value dir the writeout path
            for block in conf.writeout:
                if block.destination.resource is not None:
                    resource = block.destination.resource
                    parts = resource.split("/")
                    new_resource = os.path.join("/".join(parts[:-1]), run_id, parts[-1])
                    block.destination.resource = new_resource

            formatted_confs.append(conf)

        return formatted_confs



import os
import tempfile
from typing import Tuple, List
from copy import deepcopy

from pydantic import BaseModel

from icolos.utils.enums.step_enums import StepTurbomoleEnum
from icolos.utils.execute_external.execute import execution_successful
from icolos.utils.execute_external.openbabel import OpenBabelExecutor
from icolos.utils.execute_external.turbomole import TurbomoleExecutor
from icolos.utils.general.convenience_functions import nested_get

from icolos.utils.general.molecules import get_charge_for_molecule

from icolos.core.containers.compound import Conformer, Enumeration

from icolos.utils.enums.program_parameters import OpenBabelEnum
from icolos.utils.enums.program_parameters import TurbomoleEnum
from icolos.utils.enums.compound_enums import ConformerContainerEnum
from icolos.core.workflow_steps.calculation.base import StepCalculationBase
from icolos.core.workflow_steps.step import _LE
from icolos.loggers.logger_utils import log_multiline_string
from icolos.utils.general.files_paths import _FG, check_file_availability, gen_tmp_file

from icolos.utils.general.parallelization import Parallelizer, SubtaskContainer

_OE = OpenBabelEnum()
_EE = TurbomoleEnum()
_COE = ConformerContainerEnum()
_STE = StepTurbomoleEnum()


class StepTurbomole(StepCalculationBase, BaseModel):
    """Step that executes Turbomole.

    Note, that the execution (especially in conjunction with a subsequent cosmo step) is relatively complex.
    (1) Write the conformer as an SDF file to a temporary directory,
    (2) use obabel to translate it to an XYZ file,
    (3) use t2x to make a coord file out of it (input for turbomole; is updated during geometry optimization),
    (4) execute Turbomole, generating a (i) a final coord file and (ii) a trajectory (if specified),
    (5) use x2t to extract the final "snapshot" as an XYZ file and translate it to SDF and
    (6) update the coordinates and tags in the conformers

    IMPORTANT: Keep the "mol.cosmo" file attached to the conformer as additional data for a possible cosmo step."""

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=TurbomoleExecutor)
        # TODO: figure out why "module load turbomole/73 && rdfit" sometimes fails (see also below) and
        #       use strict=True after fix; probably, it has to to with $TURBOTMPDIR (all parallel jobs access the same)
        self._check_backend_availability(strict=False)

    def get_original_conformer(self, conformer: Conformer) -> Conformer:
        for compound in self.get_compounds():
            if (
                conformer.get_enumeration_object()._compound_object.get_compound_number()
                == compound.get_compound_number()
            ):
                for enum in compound.get_enumerations():
                    if (
                        enum._enumeration_id
                        == conformer.get_enumeration_object().get_enumeration_id()
                    ):
                        for conf in enum.get_conformers():
                            if conf._conformer_id == conformer._conformer_id:
                                return conf

    def _prepare_tmp_input_directories(
        self, batch: List
    ) -> Tuple[List, List[str], List[str], List[str], List[str], List[str], List[str]]:
        conformers = []
        tmp_dirs = []
        paths_input_sdf = []
        paths_input_xyz = []
        paths_coord = []
        paths_tm_config = []
        paths_cosmo_config = []
        for sublist in batch:
            for element in sublist:  # there is only one
                conformer = element.data
                conformers.append(conformer)
                # 1) generate all temporary paths
                tmp_dir = tempfile.mkdtemp()
                _, path_input_sdf = gen_tmp_file(
                    prefix="tmp_", suffix=".sdf", dir=tmp_dir
                )
                _, path_input_xyz = gen_tmp_file(
                    prefix="tmp_", suffix=".xyz", dir=tmp_dir
                )
                path_coord = os.path.join(tmp_dir, _EE.COORD)

                # 2) write-out the conformers for an enumeration in a SDF file
                conformer.write(path=path_input_sdf)

                # 3) translate the SDF into an XYZ file (using OpenBabel)
                # Note, that all tags are lost here (but the names are not!)
                obabel_executor = OpenBabelExecutor()
                obabel_executor.execute(
                    command=_OE.OBABEL,
                    arguments=[
                        _OE.OBABEL_INPUTFORMAT_SDF,
                        path_input_sdf,
                        _OE.OBABEL_OUTPUT_FORMAT_XYZ,
                        "".join([_OE.OBABEL_O, path_input_xyz]),
                    ],
                    check=True,
                    location=tmp_dir,
                )

                # 4) translate the XYZ to a TM input file ("coord"); "x2t" writes to stdout
                result = self._backend_executor.execute(
                    command=_EE.TM_X2T, arguments=[path_input_xyz], check=True
                )
                with open(path_coord, "w") as file:
                    file.write(result.stdout)

                # 5) add paths
                tmp_dirs.append(tmp_dir)
                paths_input_sdf.append(path_input_sdf)
                paths_input_xyz.append(path_input_xyz)
                paths_coord.append(path_coord)

                tm_path, cosmo_path = self._get_config_paths(conformer)
                paths_tm_config.append(tm_path)
                paths_cosmo_config.append(cosmo_path)

        return (
            conformers,
            tmp_dirs,
            paths_input_sdf,
            paths_input_xyz,
            paths_coord,
            paths_tm_config,
            paths_cosmo_config,
        )

    def _get_config_paths(self, conformer: Conformer) -> Tuple[str, str]:
        try:
            config_dir = self.settings.additional[_EE.TM_CONFIG_DIR]
            config_basename = self.settings.additional[_EE.TM_CONFIG_BASENAME]
            path_cosmo_config = self.settings.additional[_EE.TM_CONFIG_COSMO]
        except KeyError as e:
            raise KeyError("The dir, basename and cosmo paths need to be set.") from e

        charge = str(
            get_charge_for_molecule(
                molecule=conformer._enumeration_object.get_molecule()
            )
        )

        # the path would look like: /opt/Icolos/turbomole_config/b97-3c-ri-d3-def2-mtzvp-int-nosym-charge-1.tm
        path_tm_config = os.path.join(
            config_dir, "".join([config_basename, charge, _EE.TM_CONFIG_ENDING])
        )
        return path_tm_config, path_cosmo_config

    def _execute_define(self, tmp_dir, path_tm_config: str):
        result = self._backend_executor.execute(
            command=_EE.TM_DEFINE,
            arguments=[" ".join(["<", path_tm_config])],
            check=True,
            location=tmp_dir,
        )

        if not execution_successful(result.stderr, _EE.TM_DEFINE_SUCCESS_STRING):
            self._logger.log(
                f"Execution of {_EE.TM_DEFINE} failed for file {path_tm_config}. Error message:",
                _LE.ERROR,
            )
            log_multiline_string(
                logger=self._logger_blank,
                level=_LE.ERROR,
                multi_line_string=result.stdout,
            )

    def _execute_cosmoprep(self, tmp_dir, path_cosmo_config: str):
        result = self._backend_executor.execute(
            command=_EE.TM_COSMOPREP,
            arguments=[" ".join(["<", path_cosmo_config])],
            check=True,
            location=tmp_dir,
        )

        if not execution_successful(result.stderr, _EE.TM_COSMOPREP_SUCCESS_STRING):
            self._logger.log(
                f"Execution of {_EE.TM_COSMOPREP} failed for file {path_cosmo_config}. Error message:",
                _LE.ERROR,
            )
            log_multiline_string(
                logger=self._logger_blank,
                level=_LE.ERROR,
                multi_line_string=result.stdout,
            )

    def _manipulate_control_script(self, path: str):
        # do the following changes to the "control" script in order to generate FINE Cosmo files
        with open(path, "r") as f:
            control = f.readlines()
        new_control = []
        for line in control:
            if line.rstrip("\n") != _EE.CONTROL_COSMO_OUT:
                new_control.append(line)
            else:
                new_control.append("".join([_EE.CONTROL_COSMO_REPLACE, "\n"]))

                # only add this line in case there is no optimization run going on
                if (
                    nested_get(
                        self.settings.additional,
                        [_STE.EXECUTION_MODE],
                        default=_EE.TM_RIDFT,
                    )
                    == _EE.TM_RIDFT
                ):
                    new_control.append("".join([_EE.CONTROL_COSMO_INSERTION, "\n"]))
        with open(path, "w") as f:
            f.writelines(new_control)

    def _get_arguments(self) -> list:
        arguments = []

        # add flags
        for flag in self.settings.arguments.flags:
            arguments.append(flag)

        # flatten the dictionary into a list for command-line execution
        for key in self.settings.arguments.parameters.keys():
            arguments.append(key)
            arguments.append(self.settings.arguments.parameters[key])
        return arguments

    def _execute_run(self, tmp_dir):
        execution_mode = nested_get(
            self.settings.additional, [_STE.EXECUTION_MODE], default=_EE.TM_RIDFT
        )
        result = self._backend_executor.execute(
            command=execution_mode,
            arguments=self._get_arguments(),
            check=True,
            location=tmp_dir,
        )

        if (
            not execution_successful(result.stderr, _EE.TM_RIDFT_SUCCESS_STRING)
            or result.returncode != 0
        ):
            self._logger.log(
                f"Execution of {execution_mode} failed (return code: {result.returncode}). Error message (stdout & stderr):",
                _LE.DEBUG,
            )
            log_multiline_string(
                logger=self._logger_blank,
                level=_LE.DEBUG,
                multi_line_string=result.stdout,
            )
            log_multiline_string(
                logger=self._logger_blank,
                level=_LE.DEBUG,
                multi_line_string=result.stderr,
            )
        return result.returncode

    def _coord2sdf(self, tmp_dir, path_output_xyz: str, path_output_sdf: str):
        # extract the latest snapshot and write it as an XYZ file
        result = self._backend_executor.execute(
            command=_EE.TM_T2X, arguments=[_EE.TM_T2X_C], check=True, location=tmp_dir
        )

        with open(path_output_xyz, "w") as file:
            file.write(result.stdout)

        # translate it to an SDF
        obabel_executor = OpenBabelExecutor()
        obabel_executor.execute(
            command=_OE.OBABEL,
            arguments=[
                _OE.OBABEL_INPUTFORMAT_XYZ,
                path_output_xyz,
                _OE.OBABEL_OUTPUT_FORMAT_SDF,
                "".join([_OE.OBABEL_O, path_output_sdf]),
            ],
            check=True,
        )

    def _parse_output(self, tmp_dirs: List[str], conformers: List[Conformer]):
        results = []
        # load and attach "mol.cosmo" file
        for tmp_dir, conformer in zip(tmp_dirs, conformers):
            result = _STE.SUCCESS
            cosmo_path = os.path.join(tmp_dir, _EE.TM_OUTPUT_COSMOFILE)
            if check_file_availability(path=cosmo_path) != _FG.NOT_GENERATED:
                with open(cosmo_path, "r") as f:
                    file_content = f.readlines()
                conf = self.get_original_conformer(conformer)
                conf.add_extra_data(key=_COE.EXTRA_DATA_COSMOFILE, data=file_content)
                # conformer.add_extra_data(
                #     key=_COE.EXTRA_DATA_COSMOFILE, data=file_content
                # )

            else:
                self._logger.log(
                    f"Could not load cosmo file for {conformer.get_index_string()}, will remove conformer.",
                    _LE.WARNING,
                )
                self._logger.log(
                    f"File {cosmo_path} could not be loaded for {conformer.get_index_string()}.",
                    _LE.DEBUG,
                )
                result = _STE.FAILED

                # set molecule to None removes the 3D coordinates -> will be deleted in the end
                conformer.set_molecule(None)

            # load and attach "coord" file
            coord_file = os.path.join(tmp_dir, _EE.TM_OUTPUT_COORDFILE)
            coord_file_status = check_file_availability(path=coord_file)
            if coord_file_status == _FG.NOT_GENERATED:
                self._logger.log(
                    f"File {coord_file} could not be loaded for {conformer.get_index_string()}.",
                    _LE.DEBUG,
                )
                result = _STE.FAILED
            elif coord_file_status == _FG.GENERATED_EMPTY:
                self._logger.log(
                    f"File {coord_file} is empty for {conformer.get_index_string()}.",
                    _LE.DEBUG,
                )
                result = _STE.FAILED
            elif coord_file_status == _FG.GENERATED_SUCCESS:
                with open(coord_file, "r") as f:
                    file_content = f.readlines()
                    conf = self.get_original_conformer(conformer)
                    conf.add_extra_data(
                        key=_COE.EXTRA_DATA_COORDFILE, data=file_content
                    )

                execution_mode = nested_get(
                    self.settings.additional,
                    [_STE.EXECUTION_MODE],
                    default=_EE.TM_RIDFT,
                )

                # for RIDFT, only the cosmo file is required as coordinates are not updated (no geometry optimization)
                if execution_mode != _EE.TM_RIDFT:
                    path_output_xyz = os.path.join(tmp_dir, _EE.TM_OUTPUT_FINAL_XYZ)
                    path_output_sdf = os.path.join(tmp_dir, _EE.TM_OUTPUT_FINAL_SDF)
                    self._coord2sdf(tmp_dir, path_output_xyz, path_output_sdf)
                    conf = self.get_original_conformer(conformer)
                    conf.update_coordinates(path=path_output_sdf)
            results.append(result)
        return results

    def _clean_failed_conformers(self, enumeration: Enumeration) -> Tuple[int, int]:
        n_conformers_before = len(enumeration.get_conformers())
        enumeration.clean_failed_conformers()
        n_conformers_after = len(enumeration.get_conformers())
        return n_conformers_before, n_conformers_after

    def _run_conformer(
        self,
        conformer: Conformer,
        tmp_dir: str,
        path_tm_config: str,
        path_cosmo_config: str,
    ) -> None:
        self._execute_define(tmp_dir=tmp_dir, path_tm_config=path_tm_config)
        # execute COSMOprep (update "control")
        self._execute_cosmoprep(tmp_dir=tmp_dir, path_cosmo_config=path_cosmo_config)
        # set a necessary environment variable
        os.environ[_EE.TM_TURBOTMPDIR] = tmp_dir
        # update the "control" file
        self._manipulate_control_script(path=os.path.join(tmp_dir, _EE.CONTROL))
        # all ready; start the execution
        self._execute_run(tmp_dir)

        self._logger.log(
            f"Finished Turbomole execution for conformer {conformer.get_index_string()} in directory {tmp_dir}.",
            _LE.DEBUG,
        )

    def _execute_turbomole_parallel(self):
        parallelizer = Parallelizer(func=self._run_conformer)
        n = 1

        while self._subtask_container.done() is False:

            next_batch = self._get_sublists(
                get_first_n_lists=self._get_number_cores()
            )  # return n lists of length max_sublist_length
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            (
                conformers,
                tmp_dirs,
                paths_input_sdf,
                paths_input_xyz,
                paths_coord,
                paths_tm_configs,
                paths_cosmo_configs,
            ) = self._prepare_tmp_input_directories(next_batch)

            self._logger.log(
                f"Executing Turbomole for batch {n} containing {len(tmp_dirs)} conformers",
                _LE.INFO,
            )

            parallelizer.execute_parallel(
                conformer=conformers,
                tmp_dir=tmp_dirs,
                path_tm_config=paths_tm_configs,
                path_cosmo_config=paths_cosmo_configs,
            )

            results = self._parse_output(tmp_dirs, conformers)
            for sublist, result in zip(next_batch, results):
                # TODO: this only works if max length sublist == 1, fine for now as that is all turbomole can handle
                for task in sublist:
                    if result == _STE.SUCCESS:
                        task.set_status_success()
                    else:
                        task.set_status_failed()
            self._remove_temporary(tmp_dirs)
            n += 1

    def execute(self):
        all_conformers = []
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if self._input_object_empty(enumeration):
                    continue
                for conformer in enumeration.get_conformers():
                    # for efficient parallelisation, unroll all conformers
                    conf = deepcopy(conformer)
                    all_conformers.append(conf)

        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_conformers)
        self._execute_turbomole_parallel()



from icolos.utils.execute_external.execute import ExecutorBase
from icolos.utils.enums.program_parameters import (
    FepPlusEnum,
    SchrodingerExecutablesEnum,
)

FE = FepPlusEnum()
SEE = SchrodingerExecutablesEnum()


class FepPlusExecutor(ExecutorBase):
    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        if command not in [
            FE.FEP_MAPPER,
            FE.FEP_EXECUTOR,
            FE.FEP_ABSOLUTE_EXECUTOR,
            FE.JSC_LIST,
            FE.JSC_TAIL_FILE,
        ]:
            raise ValueError(
                "Execution command must be recognised by the executable's enum"
            )

        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=location,
            pipe_input=pipe_input,
        )

    def is_available(self):
        try:
            result = self.execute(
                command=FE.FEP_MAPPER, arguments=[FE.FEP_HELP], check=True
            )
            if FE.FEP_MAPPER_HELP_SUCCESS_STRING in result.stdout:
                return True
            return False
        except Exception as e:
            print(str(e))
            return False



#!/usr/bin/env python
#  coding=utf-8

import os
import sys
import json
import jsonschema
import argparse

from jsonschema import RefResolver

from icolos.config.schemas.schemas import construct_workflow_schema
from icolos.utils.enums.composite_agents_enums import WorkflowEnum
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.entry_points import ExecutorEnum

from icolos.utils.entry_point_functions.parsing_functions import (
    get_version_number,
    version_match,
    get_versions_as_strings,
    get_config_version_number,
)
from icolos.utils.general.citation_generator import ConsoleColours

# enums
_LE = LoggingConfigEnum()
_EE = ExecutorEnum()
_WE = WorkflowEnum()

_CC = ConsoleColours()


def main():
    reports = {_LE.INFO: 0, _LE.WARNING: 0, _LE.DEBUG: 0, _LE.ERROR: 0}

    def _report(message: str, level: str = _LE.INFO):
        levels_prefix = {
            _LE.INFO: f"{ConsoleColours.GREEN}INFO:{ConsoleColours.ENDC}",
            _LE.WARNING: f"{ConsoleColours.ORANGE}WARNING:{ConsoleColours.ENDC}",
            _LE.ERROR: f"{ConsoleColours.FAIL}{ConsoleColours.BOLD}ERROR:{ConsoleColours.ENDC}",
            _LE.DEBUG: f"{ConsoleColours.HEADER}DEBUG:{ConsoleColours.ENDC}",
        }
        reports[level] += 1
        print(levels_prefix[level], message, "\n")

    def _summary():
        print("------------------------\nValidation summary\n------------------------")
        print(f"# errors:", reports[_LE.ERROR])
        print(f"# warnings:", reports[_LE.WARNING])
        print("------------------------")

        # if an error occurred, return a failure code (otherwise, even if warnings happened, return 0)
        if reports[_LE.ERROR] > 0:
            print(
                f"\nJSON status: {ConsoleColours.FAIL}{ConsoleColours.BOLD}INVALID{ConsoleColours.ENDC}"
            )
            return 1
        else:
            print(
                f"\nJSON status: {ConsoleColours.GREEN}{ConsoleColours.BOLD}VALID{ConsoleColours.ENDC}"
            )
            return 0

    # get the input parameters and parse them
    version = get_version_number()
    parser = argparse.ArgumentParser(
        description='Implements entry point for the "Icolos validator", checking a JSON configuration file for issues.',
        epilog=f"Icolos version: {'unknown' if version is None else version}",
    )
    parser.add_argument(
        "-conf",
        type=str,
        default=None,
        help="A path to an workflow's configuration file (JSON dictionary) that is to be validated.",
    )
    args, args_unk = parser.parse_known_args()

    if args.conf is None or not os.path.isfile(args.conf):
        raise Exception(
            'Parameter "-conf" must be a relative or absolute path to a configuration (JSON) file.'
        )

    # load configuration
    with open(args.conf) as file:
        conf = file.read().replace("\r", "").replace("\n", "")
        try:
            conf = json.loads(conf)
        except json.decoder.JSONDecodeError as e:
            _report(
                f"Formatting error (missing quotation marks, invalid values, ...) in JSON, error message:\n{str(e)}. Aborting.",
                _LE.ERROR,
            )
            return _summary()

    # check version matching
    version_config = get_config_version_number(conf)
    if version_config is None:
        _report(
            f'Version number of configuration not specified, add to "header" block.',
            _LE.WARNING,
        )
    if not version_match(conf):
        version_config, version_installation = get_versions_as_strings(conf)
        _report(
            f"Versions of installation ({version_installation}) and configuration ({version_config}) do not match or are not specified.",
            _LE.WARNING,
        )

    # load the sub-schemas (e.g. for the header region and the steps etc.), construct a schema for the entire workflow
    # and validate the input JSON against it
    workflow_schema, schema_dir = construct_workflow_schema()
    schema_path = "file:///{0}/".format(schema_dir.replace("\\", "/"))
    resolver = RefResolver(schema_path, workflow_schema)
    validator = jsonschema.Draft7Validator(workflow_schema, resolver)
    for e in validator.iter_errors(conf):
        _report(str(e), level=_LE.ERROR)

    # TODO: add checks on specified paths (if not matched, give warning)
    # TODO: check global variables and settings (if not provided, give warning)

    return _summary()


if __name__ == "__main__":
    ret = main()
    sys.exit(ret)



from icolos.core.workflow_steps.step import StepBase
from pydantic import BaseModel
from icolos.utils.execute_external.rosetta import RosettaExecutor

# Wrapping for Rosetta ab initio structure prediction
# Note the execution is finicky, and requires some proper set up to get this to work
# we run everything locally, since public web servers are not an option for us.
# some config is required to get make_fragments.pl to run with all its dependencies in place


class StepRosettaAbinitio(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(data)

        self._inititalize_backend(executor=RosettaExecutor)



import os
from typing import List
from pydantic import BaseModel
from icolos.core.containers.perturbation_map import Node
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.utils.execute_external.execute import Executor
from icolos.utils.execute_external.gromacs import GromacsExecutor
from icolos.utils.general.parallelization import SubtaskContainer
from icolos.core.workflow_steps.step import _LE

_GE = GromacsEnum()
_SGE = StepGromacsEnum()

# These classes are based on the work of Vytautas Gapsys et al: https://github.com/deGrootLab/pmx/
class StepPMXSetup(StepPMXBase, BaseModel):
    """
    Create the directory tree structure.
    Requires the pmx workflow to be executing using the single_dir running mode
    Operates on the perturbation map object, runs acpype
    on the written structures to produce the amber-compatible itp files
    """

    _gromacs_executor: GromacsExecutor = None

    def __init__(self, **data):
        super().__init__(**data)
        self._gromacs_executor = GromacsExecutor(
            prefix_execution=self.execution.prefix_execution
        )
        self._initialize_backend(executor=Executor)

    def execute(self):
        # sets the number of replicas to be used throughput the pmx run
        replicas = (
            self.settings.additional["replicas"]
            if "replicas" in self.settings.additional.keys()
            else 3
        )
        if self.work_dir is None:
            self.work_dir = self._make_tmpdir()
            self._logger.log(f"Set workflow directory to {self.work_dir}", _LE.DEBUG)
        self._construct_perturbation_map(self.work_dir, replicas)
        # create the directory structure for subsequent calculations
        edges = self.get_edges()
        nodes = self.get_nodes()

        # create the input directory to sit at the top level of the workdir, contains ligands,
        # mdp and protein topology files
        os.makedirs(os.path.join(self.work_dir, "input"), exist_ok=True)
        for folder in ["ligands", "mdp", "protein"]:
            os.makedirs(os.path.join(self.work_dir, "input", folder), exist_ok=True)

        # handle protein parametrization with pdb2gmx
        protein = (
            self.get_workflow_object().workflow_data.perturbation_map.get_protein()
        )

        protein.write(os.path.join(self.work_dir, "input/protein"))

        self._parametrise_protein(protein=protein.get_file_name(), path="input/protein")

        # remove the backup file
        old_protein = [
            f
            for f in os.listdir(os.path.join(self.work_dir, "input/protein"))
            if f.endswith("#")
        ]
        # only want the parametrised processed pdb file in there
        old_protein.append(protein.get_file_name())
        for f in old_protein:
            os.remove(os.path.join(self.work_dir, "input/protein", f))

        self._clean_protein()

        mdp_dir = self.data.generic.get_argument_by_extension(
            ext="mdp", rtn_file_object=True
        )
        mdp_dir.write(os.path.join(self.work_dir, "input/mdp"))

        # parallelize the antechamber call across the pool of nodes

        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(nodes)
        self._execute_pmx_step_parallel(
            run_func=self._parametrise_nodes,
            step_id="pmx_setup",
            result_checker=self._check_results,
        )

        # create the output folder structure
        for edge in edges:
            edgepath = os.path.join(
                self.work_dir,
                str(f"{edge.node_from.get_node_hash()}_{edge.node_to.get_node_hash()}"),
            )
            hybridTopFolder = f"{edgepath}/hybridStrTop"
            os.makedirs(hybridTopFolder, exist_ok=True)

            # water/protein
            for wp in self.therm_cycle_branches:
                wppath = f"{edgepath}/{wp}"
                os.makedirs(wppath, exist_ok=True)

                # stateA/stateB
                for state in self.states:
                    statepath = f"{wppath}/{state}"
                    os.makedirs(statepath, exist_ok=True)

                    # run1/run2/run3
                    for r in range(1, replicas + 1):
                        runpath = f"{statepath}/run{r}"
                        os.makedirs(runpath, exist_ok=True)

                        # em/eq_posre/eq/transitions
                        for sim in self.sim_types:
                            simpath = f"{runpath}/{sim}".format(runpath, sim)
                            os.makedirs(simpath, exist_ok=True)

    def _check_results(self, batch: List[List[Node]]) -> List[List[bool]]:
        output_files = ["ffMOL.itp", "MOL.itp", "MOL.pdb"]
        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                subjob_results.append(
                    all(
                        [
                            os.path.isfile(
                                os.path.join(
                                    self.work_dir,
                                    "input/ligands",
                                    job.get_node_hash(),
                                    f,
                                )
                            )
                            for f in output_files
                        ]
                    )
                )
            results.append(subjob_results)
        return results



from typing import Dict, List
from icolos.core.containers.compound import Compound, Enumeration
from icolos.core.containers.perturbation_map import Edge
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
from icolos.utils.execute_external.pmx import PMXExecutor
from icolos.utils.general.parallelization import SubtaskContainer
import numpy as np
import glob
import pandas as pd
import os


class StepPMXRunAnalysis(StepPMXBase, BaseModel):
    """
    Analyses map, returns both a summary and a full results dataframe, written to top level of work_dir, and attaches properties to the compound
    """

    results_summary: pd.DataFrame = None
    results_all: pd.DataFrame = None

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=PMXExecutor)
        self.results_summary = pd.DataFrame()
        self.results_all = pd.DataFrame()

    def execute(self):

        edges = [e.get_edge_id() for e in self.get_edges()]
        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(edges)
        self._execute_pmx_step_parallel(
            run_func=self.run_analysis,
            step_id="pmx_run_analysis",
            result_checker=self._check_result,
        )
        self.analysis_summary(self.get_edges())
        # reattach compounds from perturbation map to step for writeout
        # REINVENT expects the same number of compounds back, if they failed to dock, they need to report a 0.00 score

        # discard the hub compound
        self.data.compounds = self.get_perturbation_map().compounds[1:]

        #     output_conf = edge.node_to.conformer
        #     enum: Enumeration = output_conf.get_enumeration_object()
        #     comp: Compound = enum.get_compound_object()
        #     print(enum, comp)
        #     self.data.compounds[comp.get_compound_number()].get_enumerations()[
        #         enum.get_enumeration_id()
        #     ].get_conformers()[0].set_molecule(output_conf.get_molecule())
        #     # print(comp, enum, conf)
        #     # match the output conformer to the compounds attached to the step from docking
        #     # self.data.compounds[int(comp)].get_enumerations[int(enum)].get_conformers[
        #     #     int(conf)
        #     # ] = output_conf
        #     self._logger.log(
        #         f"attached conf to step data for {output_conf.get_index_string()}",
        #         _LE.DEBUG,
        #     )

        # Edges that failed will have 0.00 attached, compounds that failed to dock and were never part of the map will get caught by the writeout method and set to 0.00

        # # the hub compound will not have data attached, this will be pruned here
        # all_confs = comp.unroll_conformers()
        # attached_prop = False
        # # check all conformers attached to compound, if ddG tag was attached, append to step compounds
        # if any(["ddG" in conf.get_molecule().GetPropNames() for conf in all_confs]):

        #     self.data.compounds.append(comp)

        # # self._logger.log(f"Failed to attach compound {comp.get_index_string()}, error was {e}", _LE.WARNING)
        # continue

    def _run_analysis_script(self, analysispath, stateApath, stateBpath):
        fA = " ".join(glob.glob("{0}/*xvg".format(stateApath)))
        fB = " ".join(glob.glob("{0}/*xvg".format(stateBpath)))
        oA = "integ0.dat"
        oB = "integ1.dat"
        wplot = "wplot.png"
        o = "results.txt"
        # TODO: at the moment we ignore flags from the command line
        # args = " ".join(self.settings.arguments.flags)

        cmd = "$PMX analyse"
        args = [
            "--quiet",
            "-fA",
            fA,
            "-fB",
            fB,
            "-o",
            o,
            "-oA",
            oA,
            "-oB",
            oB,
            "-w",
            wplot,
            "-t",
            298,
            "-b",
            100,
        ]
        self._backend_executor.execute(
            command=cmd, arguments=args, location=analysispath, check=False
        )

    def _read_neq_results(self, fname):
        try:
            with open(fname, "r") as fp:
                lines = fp.readlines()
        except FileNotFoundError:
            return
        out = []
        for l in lines:
            l = l.rstrip()
            foo = l.split()
            if "BAR: dG" in l:
                out.append(float(foo[-2]))
            elif "BAR: Std Err (bootstrap)" in l:
                out.append(float(foo[-2]))
            elif "BAR: Std Err (analytical)" in l:
                out.append(float(foo[-2]))
            elif "0->1" in l:
                out.append(int(foo[-1]))
            elif "1->0" in l:
                out.append(int(foo[-1]))
        return out

    def _fill_resultsAll(self, res, edge, wp, r):
        try:
            rowName = "{0}_{1}_{2}".format(edge, wp, r)
            self.results_all.loc[rowName, "val"] = res[2]
            self.results_all.loc[rowName, "err_analyt"] = res[3]
            self.results_all.loc[rowName, "err_boot"] = res[4]
            self.results_all.loc[rowName, "framesA"] = res[0]
            self.results_all.loc[rowName, "framesB"] = res[1]
        except IndexError:
            self._logger.log(
                f"Index Error encountered whilst parsing results to summary file for job {edge}/{wp}/{r}",
                _LE.WARNING,
            )

    def _summarize_results(self, edges: List[Edge]):
        fail_score = self._get_additional_setting("fail_score", default=100.0)
        bootnum = 1000
        for edge in edges:
            try:
                for wp in self.therm_cycle_branches:
                    dg = []
                    erra = []
                    errb = []
                    distra = []
                    distrb = []
                    for r in range(1, self.get_perturbation_map().replicas + 1):
                        rowName = "{0}_{1}_{2}".format(edge.get_edge_id(), wp, r)
                        dg.append(self.results_all.loc[rowName, "val"])
                        erra.append(self.results_all.loc[rowName, "err_analyt"])
                        errb.append(self.results_all.loc[rowName, "err_boot"])
                        distra.append(
                            np.random.normal(
                                self.results_all.loc[rowName, "val"],
                                self.results_all.loc[rowName, "err_analyt"],
                                size=bootnum,
                            )
                        )
                        distrb.append(
                            np.random.normal(
                                self.results_all.loc[rowName, "val"],
                                self.results_all.loc[rowName, "err_boot"],
                                size=bootnum,
                            )
                        )

                    rowName = "{0}_{1}".format(edge.get_edge_id(), wp)
                    distra = np.array(distra).flatten()
                    distrb = np.array(distrb).flatten()

                    if self.get_perturbation_map().replicas == 1:
                        self.results_all.loc[rowName, "val"] = dg[0]
                        self.results_all.loc[rowName, "err_analyt"] = erra[0]
                        self.results_all.loc[rowName, "err_boot"] = errb[0]
                    else:
                        self.results_all.loc[rowName, "val"] = np.mean(dg)
                        self.results_all.loc[rowName, "err_analyt"] = np.sqrt(
                            np.var(distra) / float(self.get_perturbation_map().replicas)
                        )
                        self.results_all.loc[rowName, "err_boot"] = np.sqrt(
                            np.var(distrb) / float(self.get_perturbation_map().replicas)
                        )

                # also collect self.results_summary
                rowNameWater = "{0}_{1}".format(edge.get_edge_id(), "unbound")
                rowNameProtein = "{0}_{1}".format(edge.get_edge_id(), "bound")
                dg = (
                    self.results_all.loc[rowNameProtein, "val"]
                    - self.results_all.loc[rowNameWater, "val"]
                )
                edge.ddG = dg
                try:
                    edge.node_to.get_conformer().get_molecule().SetProp("ddG", str(dg))
                    self._logger.log(
                        f"Attached score {dg} to conformer {edge.node_to.get_conformer().get_index_string()}",
                        _LE.DEBUG,
                    )
                except AttributeError as e:
                    self._logger.log(
                        f"Could not attach score to mol for edge {edge.get_edge_id()}, defaulting to {fail_score}",
                        _LE.WARNING,
                    )
                    edge.node_to.get_conformer().get_molecule().SetProp(
                        "ddG", str(fail_score)
                    )

                erra = np.sqrt(
                    np.power(self.results_all.loc[rowNameProtein, "err_analyt"], 2.0)
                    + np.power(self.results_all.loc[rowNameWater, "err_analyt"], 2.0)
                )
                edge.ddG_err = erra
                errb = np.sqrt(
                    np.power(self.results_all.loc[rowNameProtein, "err_boot"], 2.0)
                    + np.power(self.results_all.loc[rowNameWater, "err_boot"], 2.0)
                )
                rowName = edge.get_edge_id()

                self.results_summary.loc[rowName, "lig1"] = edge.get_edge_id().split(
                    "_"
                )[0]
                self.results_summary.loc[rowName, "lig2"] = edge.get_edge_id().split(
                    "_"
                )[1]
                self.results_summary.loc[rowName, "val"] = dg
                self.results_summary.loc[rowName, "err_analyt"] = erra
                self.results_summary.loc[rowName, "err_boot"] = errb

            except KeyError as e:
                print(f"Error in generating summary, error was {e}")

    def analysis_summary(self, edges: List[Edge]):
        edge_ids = [e.get_edge_id() for e in edges]
        for edge in edge_ids:
            for r in range(1, self.get_perturbation_map().replicas + 1):
                for wp in self.therm_cycle_branches:
                    analysispath = "{0}/analyse{1}".format(
                        self._get_specific_path(
                            workPath=self.work_dir, edge=edge, wp=wp
                        ),
                        r,
                    )
                    resultsfile = "{0}/results.txt".format(analysispath)
                    res = self._read_neq_results(resultsfile)
                    if res is not None:
                        self._fill_resultsAll(res, edge, wp, r)

        # the values have been collected now
        # let's calculate ddGs
        self._summarize_results(edges)
        # compare with experimental results automatically if provided
        try:
            if "exp_results" in self.settings.additional.keys() and os.path.isfile(
                self.settings.additional["exp_results"]
            ):
                exp_data = pd.read_csv(
                    self.settings.additional["exp_results"],
                    converters={"Ligand": lambda x: str(x).split(".")[0]},
                )
                # compute the experimental ddG and append to resultsSummary
                node_data = self.get_perturbation_map().node_df
                self.results_summary["exp_ddG"] = self.results_summary.apply(
                    lambda x: np.array(
                        self.compute_exp_ddG(
                            x["lig1"], x["lig2"], node_data=node_data, exp_data=exp_data
                        )
                    ),
                    axis=1,
                )
        except Exception as e:
            self._logger.log(
                f"Failed to compute experimental results, error was: {e}", _LE.WARNING
            )
        # final write to disk
        self.results_summary.to_csv(os.path.join(self.work_dir, "resultsSummary.csv"))
        self.results_all.to_csv(os.path.join(self.work_dir, "resultsAll.csv"))

    def compute_exp_ddG(
        self, lig1: str, lig2: str, node_data: pd.DataFrame, exp_data: pd.DataFrame
    ) -> float:
        """
        Compute the ddG between two ligands from experimental data
        """
        lig1_id = (
            node_data.loc[node_data["hash_id"] == lig1]["node_id"]
            .to_list()[0]
            .replace(" ", "")
        )
        lig2_id = (
            node_data.loc[node_data["hash_id"] == lig2]["node_id"]
            .to_list()[0]
            .replace(" ", "")
        )
        lig1_dG = float(
            exp_data.loc[exp_data["Ligand"] == lig1_id]["Exp. ΔG"].tolist()[0]
        )
        lig2_dG = float(
            exp_data.loc[exp_data["Ligand"] == lig2_id]["Exp. ΔG"].tolist()[0]
        )
        return lig2_dG - lig1_dG

    def run_analysis(self, jobs: List[str]):
        for idx, edge in enumerate(jobs):

            for r in range(1, self.get_perturbation_map().replicas + 1):

                # ligand
                wp = "unbound"
                analysispath = "{0}/analyse{1}".format(
                    self._get_specific_path(workPath=self.work_dir, edge=edge, wp=wp),
                    r,
                )
                os.makedirs(analysispath, exist_ok=True)
                stateApath = self._get_specific_path(
                    workPath=self.work_dir,
                    edge=edge,
                    wp=wp,
                    state="stateA",
                    r=r,
                    sim="transitions",
                )
                stateBpath = self._get_specific_path(
                    workPath=self.work_dir,
                    edge=edge,
                    wp=wp,
                    state="stateB",
                    r=r,
                    sim="transitions",
                )
                self._run_analysis_script(analysispath, stateApath, stateBpath)
                # protein
                wp = "bound"
                analysispath = "{0}/analyse{1}".format(
                    self._get_specific_path(workPath=self.work_dir, edge=edge, wp=wp),
                    r,
                )
                os.makedirs(analysispath, exist_ok=True)
                stateApath = self._get_specific_path(
                    workPath=self.work_dir,
                    edge=edge,
                    wp=wp,
                    state="stateA",
                    r=r,
                    sim="transitions",
                )
                stateBpath = self._get_specific_path(
                    workPath=self.work_dir,
                    edge=edge,
                    wp=wp,
                    state="stateB",
                    r=r,
                    sim="transitions",
                )
                self._run_analysis_script(analysispath, stateApath, stateBpath)

    def _check_result(self, batch: List[List[str]]) -> List[List[bool]]:
        """
        Look in each hybridStrTop dir and check the output pdb files exist for the edges
        """
        output_files = ["integ0.dat", "integ1.dat", "results.txt", "wplot.png"]
        analyse_folders = [
            f"analyse{i}" for i in range(1, self.get_perturbation_map().replicas + 1)
        ]
        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                subjob_results.append(
                    all(
                        [
                            os.path.isfile(
                                os.path.join(self.work_dir, job, branch, folder, f)
                            )
                            for f in output_files
                            for branch in self.therm_cycle_branches
                            for folder in analyse_folders
                        ]
                    )
                )
            results.append(subjob_results)
        return results



from typing import List

from icolos.loggers.steplogger import StepLogger
from icolos.utils.execute_external.openbabel import OpenBabelExecutor
from icolos.utils.execute_external.structcat import StructcatExecutor

from icolos.utils.enums.program_parameters import (
    OpenBabelEnum,
    SchrodingerExecutablesEnum,
)
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.general.icolos_exceptions import StepFailed

_SEE = SchrodingerExecutablesEnum()
_LE = LoggingConfigEnum()

_OE = OpenBabelEnum()


class StructcatUtil:
    def __init__(
        self,
        prefix_execution: str = None,
        binary_location: str = None,
        backend: str = _OE.OBABEL,
    ):
        self._logger = StepLogger()
        self._backend = backend
        # initialize and check executor
        if self._backend == _SEE.STRUCTCAT:
            self.executor = StructcatExecutor(
                prefix_execution=prefix_execution, binary_location=binary_location
            )
        elif self._backend == _OE.OBABEL:
            self.executor = OpenBabelExecutor()

        if not self.executor.is_available():
            raise StepFailed("Cannot initialize structcat backend - abort.")

    def concatenate(
        self,
        input_files: List[str],
        output_file: str,
        location: str = None,
    ):
        if self._backend == _SEE.STRUCTCAT:
            arguments = []
            for input_file in input_files:
                arguments = arguments + [
                    _SEE.STRUCTCAT_I,
                    input_file,
                ]
            arguments = arguments + [
                _SEE.STRUCTCAT_O,
                output_file,
            ]
            self.executor.execute(
                command=_SEE.STRUCTCAT, arguments=arguments, check=True
            )

        elif self._backend == _OE.OBABEL:
            arguments = input_files
            arguments.append("-O")
            arguments.append(output_file)
            self.executor.execute(
                command=_OE.OBABEL, arguments=arguments, check=True, location=location
            )



from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.execute_external.gromacs import GromacsExecutor
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
import os

_GE = GromacsEnum()


class StepGMXselect(StepGromacsBase, BaseModel):
    """
    Run gromacs rmsd calculation on trajectory
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def execute(self):

        tmp_dir = self._make_tmpdir()

        # write out generic files
        self.write_input_files(tmp_dir)

        flag_dict = {
            "-s": self.data.generic.get_argument_by_extension("tpr"),
            "-f": self.data.generic.get_argument_by_extension("xtc"),
            "-on": "index.ndx",
            "-select": self.settings.additional["selection"],
        }

        arguments = self._parse_arguments(flag_dict=flag_dict)
        result = self._backend_executor.execute(
            command=_GE.SELECT,
            arguments=arguments,
            location=tmp_dir,
            check=True,
        )

        self._log_result(result)
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



import os
from pydantic import BaseModel
from icolos.core.workflow_steps.schrodinger.fep_base import StepFEPBase
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.program_parameters import (
    FepPlusEnum,
    SchrodingerExecutablesEnum,
)
from icolos.utils.enums.write_out_enums import WriteOutEnum
from icolos.utils.execute_external.fep_plus import FepPlusExecutor
from rdkit import Chem

from icolos.utils.execute_external.schrodinger import SchrodingerExecutor

_WE = WriteOutEnum()
_SEE = SchrodingerExecutablesEnum()
_FE = FepPlusEnum()
_LE = LoggingConfigEnum()


class StepSchrodingerAbsoluteFEP(StepFEPBase, BaseModel):
    _schrodinger_executor: SchrodingerExecutor = None

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(executor=FepPlusExecutor)
        self._check_backend_availability()

        self._schrodinger_executor = SchrodingerExecutor(
            prefix_execution=self.execution.prefix_execution,
            binary_location=self.execution.binary_location,
        )

    def _prepare_poseviewer(self, tmp_dir: str) -> None:

        # we must have a protein structure to construct the pv by combining with existing compounds
        self.data.generic.get_argument_by_extension("pdb", rtn_file_object=True).write(
            os.path.join(tmp_dir, "receptor.pdb"), join=False
        )

        # write conformers in self.data.compound to sdf
        with Chem.SDWriter(os.path.join(tmp_dir, "compounds.sdf")) as writer:
            for compound in self.get_compounds():
                for enum in compound.get_enumerations():
                    for conf in enum.get_conformers():
                        mol = conf.get_molecule()
                        mol_name = (
                            compound.get_name()
                            if compound.get_name()
                            else compound.get_index_string()
                        )
                        mol.SetProp(_WE.RDKIT_NAME, mol_name)
                        writer.write(mol)

        # run structcat with the ligand files
        structcat_args = [
            "-ipdb",
            "receptor.pdb",
            "-isd",
            "compounds.sdf",
            "-omae",
            "abfe_pv.maegz",
        ]
        self._schrodinger_executor.execute(
            command=_SEE.STRUCTCAT,
            arguments=structcat_args,
            check=True,
            location=tmp_dir,
        )
        # schrodinger return codes not always reliable
        assert os.path.isfile(os.path.join(tmp_dir, "abfe_pv.maegz"))

    def execute(self):
        """Execute Schrodinger's ABFE workflow"""

        tmp_dir = self._make_tmpdir()
        print(tmp_dir)

        # require poseviewer file, either direct from generic data or construct from receptor and ligands from previous dock
        if self.data.generic.get_files_by_extension("maegz"):
            # poseviewer has been provided
            self.data.generic.get_argument_by_extension("maegz").write(
                os.path.join(tmp_dir, "abfe_pv.maegz"), join=False
            )
            self._logger.log("Using provided poseviewer file", _LE.DEBUG)
        else:
            self._logger.log(
                "Preparing poseviewer from receptor and compounds", _LE.DEBUG
            )
            self._prepare_poseviewer(tmp_dir)

        abfe_args = ["abfe_pv.maegz"] + self._parse_arguments()
        self._backend_executor.execute(
            command=_FE.FEP_ABSOLUTE_EXECUTOR,
            arguments=abfe_args,
            location=tmp_dir,
            check=True,
        )
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



class CheckFileGenerationEnum:

    GENERATED_SUCCESS = "generated_success"
    GENERATED_EMPTY = "generated_empty"
    NOT_GENERATED = "not_generated"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")


class JSONSchemasEnum:

    WORKFLOW_SCHEMA = "workflow"

    # sub-schemas
    HEADER_SCHEMA = "header"
    STEPS_SCHEMA = "steps"
    STEP_SCHEMA = "step"

    # try to find the internal value and return
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError

    # prohibit any attempt to set any values
    def __setattr__(self, key, value):
        raise ValueError("No changes allowed.")



import json
import os
import shutil
import string
import tempfile
from typing import Callable, List
from pydantic import BaseModel
import pandas as pd
from icolos.core.composite_agents.workflow import WorkFlow
from icolos.core.containers.compound import Compound, Conformer, Enumeration
from icolos.core.workflow_steps.step import StepBase
from sklearn.ensemble import RandomForestRegressor
import numpy as np
from icolos.core.workflow_steps.step import _LE
from icolos.utils.enums.step_enums import StepActiveLearningEnum, StepBaseEnum
from icolos.utils.enums.write_out_enums import WriteOutEnum
from icolos.utils.general.convenience_functions import nested_get
from icolos.utils.enums.step_initialization_enum import StepInitializationEnum
from rdkit import Chem
from rdkit.Chem.AllChem import (
    GetMorganFingerprintAsBitVect,
)
from sklearn.gaussian_process.kernels import DotProduct
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.ensemble import RandomForestRegressor
from icolos.core.workflow_steps.active_learning.al_utils import (
    expected_improvement,
    greedy_acquisition,
    latent_distance,
)
from modAL.models.learners import BayesianOptimizer, ActiveLearner
from modAL.acquisition import max_EI
from icolos.core.workflow_steps.active_learning.al_utils import FeedForwardNet
from skorch.regressor import NeuralNetRegressor
from skorch.callbacks import EarlyStopping
from icolos.utils.execute_external.execute import Executor
import torch
from torch import nn

_IE = StepInitializationEnum()
_SALE = StepActiveLearningEnum()
_WOE = WriteOutEnum()


class ActiveLearningBase(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    def _initialize_oracle_step_from_dict(self, step_conf: dict) -> StepBase:
        """Initialize an oracle step, note this is similar to the main workflow initialization method, but needs to be a separate initializer for the AL workflow

        :param dict step_conf: config for the step to be initialized
        :raises ValueError: Step ID must be recognised as an internal step
        :return StepBase: Initialized step
        """
        _STE = StepBaseEnum
        step_type = nested_get(step_conf, _STE.STEP_TYPE, default=None)
        step_type = None if step_type is None else step_type.upper()
        if step_type in _IE.STEP_INIT_DICT.keys():
            return _IE.STEP_INIT_DICT[step_type](**step_conf)
        else:
            raise ValueError(
                f"Backend for step {nested_get(step_conf, _STE.STEPID, '')} unknown."
            )

    def construct_fingerprints(self, library: pd.DataFrame) -> pd.DataFrame:
        """Add morgan FP column to dataframe containing rdkit mols

        :param pd.DataFrame library: lib containing the mols in column named Molecules
        :return pd.DataFrame: Modified dataframe with the extra fingerprints col
        """
        # add morgan FPs
        library[_SALE.MORGAN_FP] = library.apply(
            lambda x: np.array(
                GetMorganFingerprintAsBitVect(x[_SALE.MOLECULE], 2, nBits=2048),
                dtype=np.float32,
            ),
            axis=1,
        )
        library["fragment_id"] = [mol.GetProp("_Name") for mol in library.Molecule]

        library[_SALE.IDX] = [i for i in range(len(library))]

        return library

    def _get_acquisition_function(self) -> Callable:
        """Return the acquisition function specified in the config

        :return Callable: Functino that executes the acquisition function
        """
        acq_fn = self._get_additional_setting(
            _SALE.ACQUISITION_FUNCTION, default=_SALE.GREEDY
        )
        if acq_fn.lower() == _SALE.GREEDY:
            return greedy_acquisition
        elif acq_fn.lower() == _SALE.EI:
            return expected_improvement
        elif acq_fn.lower() == _SALE.LATENT_DISTANCE:
            return latent_distance
        else:
            raise ValueError(f"Acquisition method {acq_fn} is not supported!")

    def _initialize_learner(self) -> ActiveLearner:
        """Initialize the surrogate model specified in the config

        :raises KeyError: Unsopported model specified
        :return ActiveLearner: Initialized active learner, initialized with a surrogate model architecture and acquisition function
        """
        running_mode = self.settings.additional[_SALE.MODEL]
        acquisition_function = self._get_acquisition_function()
        if running_mode == "gpr":
            learner = BayesianOptimizer(
                estimator=GaussianProcessRegressor(
                    kernel=DotProduct(), normalize_y=True
                ),
                query_strategy=max_EI,
            )
        elif running_mode == "random_forest":
            learner = ActiveLearner(
                estimator=RandomForestRegressor(n_estimators=64),
                query_strategy=acquisition_function,
            )
        elif running_mode == "ffnn":
            device = "cuda" if torch.cuda.is_available() else "cpu"
            regressor = NeuralNetRegressor(
                FeedForwardNet,
                criterion=nn.MSELoss,
                optimizer=torch.optim.Adam,
                callbacks=[EarlyStopping(patience=20)],
                warm_start=False,
                verbose=1,
                lr=1e-4,
                device=device,
                max_epochs=500,
                batch_size=1024,
            )
            learner = ActiveLearner(
                estimator=regressor,
                query_strategy=acquisition_function,
            )
        else:
            raise KeyError(f"running mode: {running_mode} not supported")
        return learner

    def _initialize_oracle_workflow(
        self, compound_list: List[pd.Series], compound_file: str = None
    ) -> WorkFlow:
        """Initialize the oracle workflow

        :param List[pd.Series] compound_list: List of rows from the library df containing comounds to query
        :return WorkFlow: Initialized Icolos workflow
        """
        # path to the json config for the oracle config
        oracle_conf = self.settings.additional["oracle_config"]
        with open(oracle_conf, "r") as f:
            wf_config = json.load(f)

        if compound_file is None:
            # if no compound file already created (resi scanning already has the database file), create it.
            compound_file = os.path.join(self.work_dir, "compounds.sdf")
            # manually attach the compound objects to the oracle's lead step
            with Chem.SDWriter(compound_file) as writer:
                for idx, comp in enumerate(compound_list):
                    mol = comp[_SALE.MOLECULE]
                    try:
                        name = comp[_SALE.ID]
                        mol.SetProp("original_name", name)
                    except KeyError:
                        pass
                    mol.SetProp(_WOE.RDKIT_NAME, f"{idx}:0")
                    mol.SetProp(_WOE.COMPOUND_NAME, f"{idx}:0")
                    writer.write(mol)

        compound_dict = {
            "source": compound_file,
            "source_type": "file",
            "format": "SDF",
        }
        try:
            # if other compound are already specified, try to add another entry to the list
            wf_config["workflow"]["steps"][0]["input"]["compounds"].append(
                compound_dict
            )
        except KeyError:
            try:
                # if an input block is present but no compounds block is there
                wf_config["workflow"]["steps"][0]["input"]["compounds"] = [
                    compound_dict
                ]
            except KeyError:
                # no input block present in json, no risk of overwriting existing compounds
                wf_config["workflow"]["steps"][0]["input"] = {}
                wf_config["workflow"]["steps"][0]["input"]["compounds"] = [
                    compound_dict
                ]
        # inherit header from main workflow
        header = self.get_workflow_object().header
        oracle_wf = WorkFlow(**wf_config["workflow"])
        oracle_wf.header = header
        oracle_steps = []
        for step_conf in oracle_wf.steps:
            step_conf = oracle_wf._update_global_variables(conf=step_conf)
            st = self._initialize_oracle_step_from_dict(step_conf)

            st.set_workflow_object(oracle_wf)
            oracle_steps.append(st)

        for step in oracle_steps:
            oracle_wf.add_step(step)
        return oracle_wf

    def _run_oracle_wf(self, oracle_wf: WorkFlow, work_dir: str = None) -> WorkFlow:
        """Run the oracle workflow through for one set of compounds

        :param WorkFlow oracle_wf: initialized oracle workflow to be run
        :param str work_dir: _description_, defaults to None
        :return WorkFlow: Return completed workflow with data attached
        """
        for idx, step in enumerate(oracle_wf._initialized_steps):
            step.generate_input()
            self._logger.log(
                f"Starting execution of oracle step: {step.step_id}", _LE.INFO
            )
            if work_dir is not None:
                step.work_dir = work_dir
            step.execute()
            self._logger.log(
                f"Processing write-out blocks for {step.step_id}.", _LE.DEBUG
            )
            step.process_write_out()

        return oracle_wf

    def query_oracle(
        self,
        compound_list: List[pd.Series],
        oracle_type: str = "docking",
        fragment_lib: pd.DataFrame = None,
        round: int = 0,
    ) -> np.ndarray:
        """Main interface method with the oracle, controls initialization and executino

        :param List[pd.Series] compound_list: list of rows from library df to be queried by the oracle
        :param str oracle_type: type of oracle to initialize, defaults to "docking"
        :param pd.DataFrame fragment_lib: optional fragment library for fep and resi scanning oracles, defaults to None
        :raises NotImplementedError: Handles unknown oracle types
        :return np.ndarray: return array of scores for each compound from the oracle
        """
        criteria = self._get_additional_setting(_SALE.CRITERIA)
        if oracle_type == "pmx_rbfe":
            # test code, append the right tags to the oracle compounds

            # output_compounds = []
            # for i in range(len(compound_list)):
            #     # simulate some dropped conformers
            #     if i not in (2, 4):
            #         parent_compound = Compound(name=f"{i}", compound_number=i)
            #         parent_enum = Enumeration(enumeration_id=0)
            #         parent_enum.set_compound_object(parent_compound)
            #         conf = Conformer(Chem.MolFromSmiles("CCCC"), conformer_id=0)
            #         conf.get_molecule().SetProp(criteria, str(np.random.uniform(-5, 5)))
            #         conf.set_enumeration_object(parent_enum)
            #         output_compounds.append(conf)

            oracle_wf = self._initialize_oracle_workflow(compound_list)
            # # # we have a fully initialized step with the compounds loaded.  Execute them
            oracle_wf = self._run_oracle_wf(oracle_wf=oracle_wf)

            output_compounds = [
                n.conformer
                for n in oracle_wf._initialized_steps[-1]
                .get_perturbation_map()
                .get_nodes()
            ]
            original_compounds = [
                mol
                for mol in Chem.SDMolSupplier(
                    os.path.join(self.work_dir, "compounds.sdf")
                )
            ]
            final_compounds = []
            for comp in original_compounds:
                # look through the final compounds and find the one with the right comp/enumeration
                comp_name = comp.GetProp(_WOE.RDKIT_NAME)
                match_found = False
                for out_comp in output_compounds:
                    if (
                        comp_name
                        == out_comp.get_enumeration_object().get_index_string()
                    ):
                        final_compounds.append(out_comp)
                        self._logger.log(
                            f"matched output conformer for mol {comp_name}", _LE.DEBUG
                        )
                        match_found = True
                if not match_found:
                    # the enumeration was dropped in the workflow, append the start compound with no tags, will be zeroed
                    final_compounds.append(Conformer(conformer=comp))
                    self._logger.log(
                        f"No conformer belonging to enumeration {comp_name} found!",
                        _LE.WARNING,
                    )

            # TODO: really this should be coupled to the oracle type and should not change
            final_scores = []
            for conf in final_compounds:
                try:
                    final_scores.append(float(conf.get_molecule().GetProp(criteria)))
                except KeyError:
                    self._logger.log(
                        f"FEP score was not attached to conformer!", _LE.WARNING
                    )
                    final_scores.append(0.0)
            # move the old dir to backup
            try:
                shutil.move("output", f"output_{round}")
                os.makedirs("output")
            except Exception as e:
                print("Could not back up output files!, error was ", e)
            return np.array(final_scores, dtype=np.float32)
        elif oracle_type == "docking":

            oracle_wf = self._initialize_oracle_workflow(compound_list)
            # we have a fully initialized step with the compounds loaded.  Execute them
            oracle_wf = self._run_oracle_wf(oracle_wf=oracle_wf)

            final_compounds = oracle_wf._initialized_steps[-1].data.compounds

        elif oracle_type in ("protein_FEP", "residue_scanning"):
            # do not pass scores, generate a tmpdir, create the NCAA database, create the mutations file and run the FEP job on AWS
            # create tmpdir
            orig_dir = os.getcwd()
            tmp_dir = tempfile.mkdtemp()
            os.chdir(tmp_dir)

            # extract the relevant amino acids using compound indices from the ncaa library
            fragment_ids = [row.ID for row in compound_list]
            # retrieve the fragment using the index of the enumerated compound
            # frags = [fragment_lib.iloc[idx] for idx in compound_index]
            print("fragment ids", fragment_ids)
            frags = []
            for frag_id in fragment_ids:
                for _, frag in fragment_lib.iterrows():
                    if frag.fragment_id == frag_id:
                        frags.append(frag.Molecule)
                        print(f"found frag {frag_id}")
                        break
            letter_strings = string.ascii_uppercase + "0123456789"
            all_letters = []
            for char1 in letter_strings:
                for char2 in letter_strings:
                    for char3 in letter_strings:
                        all_letters.append(f"{char1}{char2}{char3}")
            line_stub = self._get_additional_setting("mut_res")
            # create mutations file simultaneously
            with open("mutations.txt", "w") as f, Chem.SDWriter(
                "compounds.sdf"
            ) as writer:
                for idx, frag in enumerate(frags):
                    # rename the molecule to align with what biolumiate will call i
                    frag.SetProp(_WOE.RDKIT_NAME, f"{all_letters[idx]}")
                    writer.write(frag)
                    f.write(f"{line_stub}->{all_letters[idx]}\n")
                    idx += 1

            command = "$SCHRODINGER/run python3 $NSR_LIB_SCRIPT compounds.sdf"

            executor = Executor(prefix_execution="ml schrodinger")
            executor.execute(command, arguments=[], check=True, location=tmp_dir)
            # now ncaa_nca.maegz will be in the tmpdir

            # execution is the same for both cases, just different oracle
            oracle_wf = self._initialize_oracle_workflow(
                compound_list=compound_list,
                compound_file=os.path.join(tmp_dir, "compounds.sdf"),
            )
            oracle_wf = self._run_oracle_wf(oracle_wf=oracle_wf, work_dir=tmp_dir)
            final_compounds = oracle_wf._initialized_steps[-1].data.compounds
            os.chdir(orig_dir)

        else:
            raise NotImplementedError(f"Oracle type {oracle_type} not implemented")
        self._logger.log(
            f"Extracting final scores for {len(final_compounds)} compounds", _LE.DEBUG
        )

        scores = self._extract_final_scores_from_compounds(
            final_compounds, self.settings.additional[_SALE.CRITERIA]
        )
        return scores

    def _extract_final_scores_from_compounds(
        self, compounds: List[Compound], criteria: str, highest_is_best: bool = False
    ) -> np.ndarray:
        """Extract final scores from compounds: attempts first search on the enumerations, falling back to conformers

        :param List[Compound] compounds: compounds extracted from final step in workflow
        :param str criteria: tag to extract score by
        :param bool highest_is_best: control whether highest vals correspond to best scores, negative for docking scores, affinities etc, defaults to False
        :return np.ndarray: array containing final scores per compound
        """
        top_scores = []
        for comp in compounds:
            # extract top score per compound
            scores = []
            for enum in comp.get_enumerations():
                # look in the enumeratino's mol first
                try:

                    score = enum.get_molecule().GetProp(criteria)
                    scores.append(score)
                    self._logger.log(
                        f"Got score: {score} for enum {enum.get_index_string()}",
                        _LE.DEBUG,
                    )
                except KeyError:
                    # no score attached to enumeration, search through conformers
                    for conf in enum.get_conformers():
                        try:
                            conf_score = conf.get_molecule().GetProp(criteria)
                            scores.append(float(conf_score))
                        except KeyError:
                            self._logger.log(
                                "Property not attached to either enum or conformer, score is 0.0",
                                _LE.WARNING,
                            )
                            scores.append(0.0)
            # if no enumeration attached to the compounds i.e. ligprep failed
            if not scores:
                scores.append(0.0)

            best_score = max(scores) if highest_is_best else min(scores)
            top_scores.append(best_score)

        return np.array(top_scores, dtype=np.float32)



import pandas as pd
from typing import List
from pydantic import BaseModel

from icolos.core.containers.compound import Conformer

from icolos.utils.enums.step_enums import StepRMSFilterEnum
from icolos.core.workflow_steps.step import _LE
from icolos.core.workflow_steps.calculation.base import StepCalculationBase

_SRF = StepRMSFilterEnum()


class StepRMSFilter(StepCalculationBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # extend parameters
        if _SRF.THRESHOLD not in self.settings.additional.keys():
            self.settings.additional[_SRF.THRESHOLD] = 1
        if _SRF.METHOD not in self.settings.additional.keys():
            self.settings.additional[_SRF.METHOD] = _SRF.METHOD_ALIGNMOL
        if _SRF.ORDER_BY not in self.settings.additional.keys():
            self.settings.additional[_SRF.ORDER_BY] = None
        else:
            if _SRF.ORDER_ASCENDING not in self.settings.additional.keys():
                self._logger.log(
                    'Setting order ascending not specified, setting to "True" (default).',
                    _LE.WARNING,
                )
                self.settings.additional[_SRF.ORDER_ASCENDING] = False

    def _get_representative_indices(
        self, df_rms: pd.DataFrame, prop_values: List[float]
    ) -> List[int]:
        keep_indices = []
        prop_idx = list(zip(prop_values, list(range(len(prop_values)))))
        threshold = self.settings.additional[_SRF.THRESHOLD]
        while len(prop_idx) > 0:
            # get the best (according to the property) element's index, add it to the list and remove it from
            # the remaining ones
            if self.settings.additional[_SRF.ORDER_BY] is not None:
                prop_idx = [
                    (prop, idx)
                    for prop, idx in sorted(
                        prop_idx, reverse=self.settings.additional[_SRF.ORDER_ASCENDING]
                    )
                ]
            cur_best_idx = prop_idx[0][1]
            keep_indices.append(cur_best_idx)
            del prop_idx[0]

            # remove all, that are fulfilling the RMS threshold
            for i in reversed(range(len(prop_idx))):
                comp_idx = prop_idx[i][1]
                cur_rms = df_rms.iloc[cur_best_idx, comp_idx]
                if cur_rms <= threshold:
                    del prop_idx[i]
        return keep_indices

    def _filter_conformers(self, conformers: List[Conformer]) -> List[Conformer]:
        # to select the "best" conformers, here the property to use for ordering / ranking is specified
        order_by = self.settings.additional[_SRF.ORDER_BY]
        if order_by is not None:
            prop_values = self._get_property_values(conformers, order_by)
        else:
            prop_values = [None for _ in range(len(conformers))]

        # generate RMS matrix (NxN, where N is the number of conformers)
        df_rms = self._calculate_rms_matrix(conformers, self._get_rms_method())

        keep_indices = self._get_representative_indices(df_rms, prop_values)

        return [conformers[i] for i in range(len(conformers)) if i in keep_indices]

    def execute(self):
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if self._input_object_empty(enumeration):
                    continue

                number_conformers_before = len(enumeration)

                # filter conformers on the enumeration level
                filtered_conformers = self._filter_conformers(
                    conformers=enumeration.get_conformers()
                )

                # add filtered conformers to enumeration
                enumeration.clear_conformers()
                for conf in filtered_conformers:
                    enumeration.add_conformer(conformer=conf, auto_update=True)
                number_conformers_after = len(enumeration)
                self._logger.log(
                    f"Filtered {number_conformers_before} conformers down to {number_conformers_after} for enumeration {enumeration.get_index_string()}.",
                    _LE.INFO,
                )



def stringify(obj):
    """Converts all objects in a dict to str, recursively."""
    if isinstance(obj, dict):
        return {str(key): stringify(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [stringify(value) for value in obj]
    else:
        return str(obj)



from copy import deepcopy

from icolos.core.containers.compound import Enumeration, Compound, Conformer
from icolos.utils.enums.write_out_enums import RunVariablesEnum

_RVE = RunVariablesEnum()


class RunVariablesResolver:
    def __init__(self):
        pass

    def _replace(self, input_str: str, pattern: str, replacement) -> str:
        if replacement is not None:
            pattern = _RVE.PREFIX + pattern + _RVE.POSTFIX
            input_str = input_str.replace(pattern, str(replacement))
        return input_str

    def resolve_compound_level(self, input_str: str, comp: Compound) -> str:
        comp = deepcopy(comp)
        resolved_str = self._replace(
            input_str, _RVE.COMPOUND_ID, comp.get_compound_number()
        )
        resolved_str = self._replace(resolved_str, _RVE.COMPOUND_NAME, comp.get_name())
        return resolved_str

    def resolve_enumeration_level(self, input_str: str, enum: Enumeration) -> str:
        enum = deepcopy(enum)
        resolved_str = self._replace(
            input_str, _RVE.ENUMERATION_ID, enum.get_enumeration_id()
        )
        resolved_str = self._replace(
            resolved_str, _RVE.ENUMERATION_STRING, enum.get_index_string()
        )
        return resolved_str

    def resolve_conformer_level(self, input_str: str, conf: Conformer) -> str:
        conf = deepcopy(conf)
        resolved_str = self._replace(
            input_str, _RVE.CONFORMER_ID, conf.get_conformer_id()
        )
        resolved_str = self._replace(
            resolved_str, _RVE.CONFORMER_STRING, conf.get_index_string()
        )
        return resolved_str

    def resolve(self, input_str: str, input_object) -> str:
        if not isinstance(input_str, str):
            return input_str

        if isinstance(input_object, Conformer):
            input_str = self.resolve_compound_level(
                input_str, input_object.get_enumeration_object().get_compound_object()
            )
            input_str = self.resolve_enumeration_level(
                input_str, input_object.get_enumeration_object()
            )
            return self.resolve_conformer_level(input_str, input_object)
        elif isinstance(input_object, Enumeration):
            input_str = self.resolve_compound_level(
                input_str, input_object.get_compound_object()
            )
            return self.resolve_enumeration_level(input_str, input_object)
        elif isinstance(input_object, Compound):
            return self.resolve_compound_level(input_str, input_object)
        else:
            raise ValueError(f'Object of type "{type(input_object)}" is not supported.')



from typing import Callable, List
from pydantic import BaseModel
import numpy as np
from icolos.core.workflow_steps.step import _LE


class ActiveLearner(BaseModel):
    """
    Implements the base active learner
    :param fit_to_new: bool: control whether to train the model just on the new data, or on the entire dataset at each refitting step
    """

    def __init__(
        self,
        estimator,
        acquisition_function: Callable,
        fit_to_new: bool = True,
        X: np.ndarray = None,
        y: np.ndarray = None,
    ) -> None:
        self.estimator = estimator
        self.acquisition_function = acquisition_function
        self.fit_to_new = fit_to_new
        self.X = X
        self.y = y

    def teach(self, X: np.ndarray, y: np.ndarray):
        """
        Take new batch of data from the oracle, train the model
        """
        if self.fit_to_new:
            # just fit the model on the new data, works for tf models etc
            self.estimator.fit(X, y)
        else:
            self.append_new_data(X, y)
            self.estimator.fit(self.X, self.y)

    def append_new_data(self, X: np.ndarray, y: np.ndarray):
        if self.X is None:
            self.X = X
        else:
            # TODO: make sure dims match here
            self.X = np.concatenate(self.X, self)
        if self.y is None:
            self.y = y
        else:
            # TODO: make sure dims match here
            self.y = np.concatenate(self.y, self)

    def query(self, previous_idx: List[int], n_samples: int):
        """
        Query the model using the acquisition function, return the n_samples
        :param previous_idx: list of indices of previously requested samples
        """
        return self.acquisition_function(
            self.estimator, self.X, previous_idx, n_samples
        )



class ExecutionFailed(Exception):
    pass


class StepFailed(Exception):
    pass


class ContainerCorrupted(Exception):
    pass


def get_exception_message(e: Exception):
    if e is None:
        return None
    if hasattr(e, "message"):
        return e.message
    else:
        return e


def get_exception_type(e: Exception) -> str:
    if e is None:
        return None
    return type(e).__name__
